import json
import typing
from http import cookies as http_cookies
from urllib.parse import parse_qs

import anyio

from ..utils._utils import AwaitableOrContextManager, AwaitableOrContextManagerWrapper
from ..datastructure.core import URL, Address, FormData, Headers, State
from ..exception.http_exception import HTTPException
from ..datastructure.formparser import FormParser, MultiPartException, MultiPartParser
from ..datastructure.user_agent import UserAgentParser
from ..types import Message, Receive, Scope, Send

try:
    from multipart.multipart import parse_options_header
except ModuleNotFoundError: 
    parse_options_header = None


SERVER_PUSH_HEADERS_TO_COPY = {
    "accept",
    "accept-encoding",
    "accept-language",
    "cache-control",
    "user-agent",
}


def cookie_parser(cookie_string: str) -> typing.Dict[str, str]:
    cookie_dict: typing.Dict[str, str] = {}
    for chunk in cookie_string.split(";"):
        if "=" in chunk:
            key, val = chunk.split("=", 1)
        else:
            key, val = "", chunk
        key, val = key.strip(), val.strip()
        if key or val:
            cookie_dict[key] = http_cookies._unquote(val)
    return cookie_dict


class ClientDisconnect(Exception):
    pass


class HTTPConnection(typing.Mapping[str, typing.Any]):

    def __init__(self, scope: Scope, receive: typing.Optional[Receive] = None) -> None:
        assert scope["type"] in ("http", "websocket")
        self.scope = scope

    def __getitem__(self, key: str) -> typing.Any:
        return self.scope[key]

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self.scope)

    def __len__(self) -> int:
        return len(self.scope)
    
    __eq__ = object.__eq__ 
    __hash__ = object.__hash__

    @property
    def app(self) -> typing.Any:
        return self.scope["app"]

    @property
    def url(self) -> URL:
        if not hasattr(self, "_url"):
            self._url = URL(scope=self.scope)
        return self._url

    @property
    def base_url(self) -> URL:
        if not hasattr(self, "_base_url"):
            base_url_scope = dict(self.scope)
            base_url_scope["path"] = "/"
            base_url_scope["query_string"] = b""
            base_url_scope["root_path"] = base_url_scope.get(
                "app_root_path", base_url_scope.get("root_path", "")
            )
            self._base_url = URL(scope=base_url_scope)
        return self._base_url

    @property
    def headers(self) -> Headers:
        if not hasattr(self, "_headers"):
            self._headers = Headers(scope=self.scope)
        return self._headers
    
    @property
    def origin(self):
        return self.headers.get('origin')
    
    @property
    def remote_addr(self) -> str:
        return self.scope.get('client', ('',))[0]
    
    @property
    def scheme(self) -> str:
        return self.scope.get('scheme', 'http')
    
    @property
    def server(self) -> typing.Dict[str, str]:
        return {
            'server_protocol': self.scope.get('server_protocol', ''),
            'server_name': self.scope.get('server_name', ''),
            'server_port': self.scope.get('server_port', ''),
        }
    
    @property
    def authorization(self) -> typing.Optional[str]:
        return self.headers.get('authorization')
    
    @property
    def user_agent(self) -> UserAgentParser:
        return UserAgentParser(self.headers.get('user-agent', ''))

    @property
    def referer(self) -> str:
        return self.headers.get('referer', '')

    @property
    def accept(self) -> str:
        return self.headers.get('accept', '')
    
    @property
    def host(self) -> str:
        return self.headers.get('host')
    
    @property
    def path(self) -> str:
        return self.scope.get('path', '/')

    @property
    def path_param(self) -> typing.Dict[str, typing.Any]:
        return self.scope.get("path_params", {})

    @property
    def cookies(self) -> typing.Dict[str, str]:
        if not hasattr(self, "_cookies"):
            cookies: typing.Dict[str, str] = {}
            cookie_header = self.headers.get("cookie")

            if cookie_header:
                cookies = cookie_parser(cookie_header)
            self._cookies = cookies
        return self._cookies

    @property
    def client(self) -> typing.Optional[Address]:
        # client is a 2 item tuple of (host, port), None or missing
        host_port = self.scope.get("client")
        if host_port is not None:
            return Address(*host_port)
        return None

    @property
    def session(self) -> typing.Dict[str, typing.Any]:
        assert (
            "session" in self.scope
        ), "SessionMiddleware must be installed to access request.session"
        return self.scope["session"]

    @property
    def auth(self) -> typing.Any:
        assert (
            "auth" in self.scope
        ), "AuthenticationMiddleware must be installed to access request.auth"
        return self.scope["auth"]

    @property
    def user(self) -> typing.Any:
        assert (
            "user" in self.scope
        ), "AuthenticationMiddleware must be installed to access request.user"
        return self.scope["user"]

    @property
    def state(self) -> State:
        if not hasattr(self, "_state"):
            self.scope.setdefault("state", {})
            self._state = State(self.scope["state"])
        return self._state

async def empty_receive() -> typing.NoReturn:
    raise RuntimeError("Receive channel has not been made available")


async def empty_send(message: Message) -> typing.NoReturn:
    raise RuntimeError("Send channel has not been made available")


class Request(HTTPConnection):
    _form: typing.Optional[FormData]

    def __init__(
        self, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send
    ):
        super().__init__(scope)
        assert scope["type"] == "http"
        self._receive = receive
        self._send = send
        self._stream_consumed = False
        self._is_disconnected = False
        self._form = None
        self.path_params = None
        self.query_params: typing.Dict[str, str] = {}
        self.context: typing.Dict[str, str] = {}
        self.executed_middlewares = set()

    @property
    def method(self) -> str:
        return self.scope["method"]
    
    @property
    def args(self) -> typing.Dict[str, str]:
        self._parse_query_params()
        return self.query_params

    @property
    def receive(self) -> Receive:
        return self._receive

    async def stream(self) -> typing.AsyncGenerator[bytes, None]:
        if hasattr(self, "_body"):
            yield self._body
            yield b""
            return
        if self._stream_consumed:
            raise RuntimeError("Stream consumed")
        self._stream_consumed = True
        while True:
            message = await self._receive()
            if message["type"] == "http.request":
                body = message.get("body", b"")
                if body:
                    yield body
                if not message.get("more_body", False):
                    break
            elif message["type"] == "http.disconnect":
                self._is_disconnected = True
        yield b""

    async def body(self) -> bytes:
        if not hasattr(self, "_body"):
            chunks: "typing.List[bytes]" = []
            async for chunk in self.stream():
                chunks.append(chunk)
            self._body = b"".join(chunks)
        return self._body

    async def json(self) -> typing.Any:
        if not hasattr(self, "_json"):
            body = await self.body()
            self._json = json.loads(body)
        return self._json
    
    def _parse_query_params(self):
        query_string = self.scope.get('query_string', b'').decode('utf-8')
        self.query_params = {k: v[0] for k, v in parse_qs(query_string).items()}

    async def _get_form(
        self,
        *,
        max_files: typing.Union[int, float] = 1000,
        max_fields: typing.Union[int, float] = 1000,
    ) -> FormData:
        if self._form is None:
            assert (
                parse_options_header is not None
            ), "The `python-multipart` library must be installed to use form parsing."
            content_type_header = self.headers.get("Content-Type")
            content_type: bytes
            content_type, _ = parse_options_header(content_type_header)
            if content_type == b"multipart/form-data":
                try:
                    multipart_parser = MultiPartParser(
                        self.headers,
                        self.stream(),
                        max_files=max_files,
                        max_fields=max_fields,
                    )
                    self._form = await multipart_parser.parse()
                except MultiPartException as exc:
                    if "app" in self.scope:
                        raise HTTPException(status_code=400, detail=exc.message)
                    raise exc
            elif content_type == b"application/x-www-form-urlencoded":
                form_parser = FormParser(self.headers, self.stream())
                self._form = await form_parser.parse()
            else:
                self._form = FormData()
        return self._form

    def form(
        self,
        *,
        max_files: typing.Union[int, float] = 1000,
        max_fields: typing.Union[int, float] = 1000,
    ) -> AwaitableOrContextManager[FormData]:
        return AwaitableOrContextManagerWrapper(
            self._get_form(max_files=max_files, max_fields=max_fields)
        )

    async def close(self) -> None:
        if self._form is not None:
            await self._form.close()

    async def is_disconnected(self) -> bool:
        if not self._is_disconnected:
            message: Message = {}

            with anyio.CancelScope() as cs:
                cs.cancel()
                message = await self._receive()

            if message.get("type") == "http.disconnect":
                self._is_disconnected = True

        return self._is_disconnected

    async def send_push_promise(self, path: str) -> None:
        if "http.response.push" in self.scope.get("extensions", {}):
            raw_headers: "typing.List[typing.Tuple[bytes, bytes]]" = []
            for name in SERVER_PUSH_HEADERS_TO_COPY:
                for value in self.headers.getlist(name):
                    raw_headers.append(
                        (name.encode("latin-1"), value.encode("latin-1"))
                    )
            await self._send(
                {"type": "http.response.push", "path": path, "headers": raw_headers}
            )




import base64
import datetime
import json
import time
import warnings
import zlib

from aquilify.settings import settings
from aquilify.utils.crypto import constant_time_compare, salted_hmac
from aquilify.utils.encoding import force_bytes
from aquilify.utils.module_loading import import_string
from aquilify.utils.regex_helper import _lazy_re_compile

_SEP_UNSAFE = _lazy_re_compile(r"^[A-z0-9-_=]*$")
BASE62_ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"


class BadSignature(Exception):
    """Signature does not match."""

    pass


class SignatureExpired(BadSignature):
    """Signature timestamp is older than required max_age."""

    pass


def b62_encode(s):
    if s == 0:
        return "0"
    sign = "-" if s < 0 else ""
    s = abs(s)
    encoded = ""
    while s > 0:
        s, remainder = divmod(s, 62)
        encoded = BASE62_ALPHABET[remainder] + encoded
    return sign + encoded


def b62_decode(s):
    if s == "0":
        return 0
    sign = 1
    if s[0] == "-":
        s = s[1:]
        sign = -1
    decoded = 0
    for digit in s:
        decoded = decoded * 62 + BASE62_ALPHABET.index(digit)
    return sign * decoded


def b64_encode(s):
    return base64.urlsafe_b64encode(s).strip(b"=")


def b64_decode(s):
    pad = b"=" * (-len(s) % 4)
    return base64.urlsafe_b64decode(s + pad)


def base64_hmac(salt, value, key, algorithm="sha1"):
    return b64_encode(
        salted_hmac(salt, value, key, algorithm=algorithm).digest()
    ).decode()


def _cookie_signer_key(key):
    return b"aquilify.http.cookies" + force_bytes(key)


def get_cookie_signer(salt="aquilify.core.signing.get_cookie_signer"):
    Signer = import_string(settings.SIGNING_BACKEND)
    return Signer(
        key=_cookie_signer_key(settings.SECRET_KEY),
        fallback_keys=map(_cookie_signer_key, settings.SECRET_KEY_FALLBACKS),
        salt=salt,
    )


class JSONSerializer:

    def dumps(self, obj):
        return json.dumps(obj, separators=(",", ":")).encode("latin-1")

    def loads(self, data):
        return json.loads(data.decode("latin-1"))


def dumps(
    obj, key=None, salt="aquilify.core.signing", serializer=JSONSerializer, compress=False
):
    return TimestampSigner(key=key, salt=salt).sign_object(
        obj, serializer=serializer, compress=compress
    )


def loads(
    s,
    key=None,
    salt="aquilify.core.signing",
    serializer=JSONSerializer,
    max_age=None,
    fallback_keys=None,
):
    """
    Reverse of dumps(), raise BadSignature if signature fails.

    The serializer is expected to accept a bytestring.
    """
    return TimestampSigner(
        key=key, salt=salt, fallback_keys=fallback_keys
    ).unsign_object(
        s,
        serializer=serializer,
        max_age=max_age,
    )


class Signer:
    def __init__(
        self,
        *args,
        key=None,
        sep=":",
        salt=None,
        algorithm=None,
        fallback_keys=None,
    ):
        self.key = key or settings.SECRET_KEY
        self.fallback_keys = (
            fallback_keys
            if fallback_keys is not None
            else settings.SECRET_KEY_FALLBACKS
        )
        self.sep = sep
        self.salt = salt or "%s.%s" % (
            self.__class__.__module__,
            self.__class__.__name__,
        )
        self.algorithm = algorithm or "sha256"
        if args:
            warnings.warn(
                f"Passing positional arguments to {self.__class__.__name__} is "
                f"deprecated.",
                stacklevel=2,
            )
            for arg, attr in zip(
                args, ["key", "sep", "salt", "algorithm", "fallback_keys"]
            ):
                if arg or attr == "sep":
                    setattr(self, attr, arg)
        if _SEP_UNSAFE.match(self.sep):
            raise ValueError(
                "Unsafe Signer separator: %r (cannot be empty or consist of "
                "only A-z0-9-_=)" % sep,
            )

    def signature(self, value, key=None):
        key = key or self.key
        return base64_hmac(self.salt + "signer", value, key, algorithm=self.algorithm)

    def sign(self, value):
        return "%s%s%s" % (value, self.sep, self.signature(value))

    def unsign(self, signed_value):
        if self.sep not in signed_value:
            raise BadSignature('No "%s" found in value' % self.sep)
        value, sig = signed_value.rsplit(self.sep, 1)
        for key in [self.key, *self.fallback_keys]:
            if constant_time_compare(sig, self.signature(value, key)):
                return value
        raise BadSignature('Signature "%s" does not match' % sig)

    def sign_object(self, obj, serializer=JSONSerializer, compress=False):
        data = serializer().dumps(obj)
        is_compressed = False

        if compress:
            compressed = zlib.compress(data)
            if len(compressed) < (len(data) - 1):
                data = compressed
                is_compressed = True
        base64d = b64_encode(data).decode()
        if is_compressed:
            base64d = "." + base64d
        return self.sign(base64d)

    def unsign_object(self, signed_obj, serializer=JSONSerializer, **kwargs):
        base64d = self.unsign(signed_obj, **kwargs).encode()
        decompress = base64d[:1] == b"."
        if decompress:
            base64d = base64d[1:]
        data = b64_decode(base64d)
        if decompress:
            data = zlib.decompress(data)
        return serializer().loads(data)


class TimestampSigner(Signer):
    def timestamp(self):
        return b62_encode(int(time.time()))

    def sign(self, value):
        value = "%s%s%s" % (value, self.sep, self.timestamp())
        return super().sign(value)

    def unsign(self, value, max_age=None):
        result = super().unsign(value)
        value, timestamp = result.rsplit(self.sep, 1)
        timestamp = b62_decode(timestamp)
        if max_age is not None:
            if isinstance(max_age, datetime.timedelta):
                max_age = max_age.total_seconds()
            age = time.time() - timestamp
            if age > max_age:
                raise SignatureExpired("Signature age %s > %s seconds" % (age, max_age))
        return value



import binascii
import json

from aquilify.settings import settings
from .base import BaseStorage, Message
from aquilify.core import signing
from aquilify.core.http.cookie import SimpleCookie
from aquilify.utils.safestring import SafeData, mark_safe


class MessageEncoder(json.JSONEncoder):

    message_key = "__json_message"

    def default(self, obj):
        if isinstance(obj, Message):
            is_safedata = 1 if isinstance(obj.message, SafeData) else 0
            message = [self.message_key, is_safedata, obj.level, obj.message]
            if obj.extra_tags is not None:
                message.append(obj.extra_tags)
            return message
        return super().default(obj)


class MessageDecoder(json.JSONDecoder):
    def process_messages(self, obj):
        if isinstance(obj, list) and obj:
            if obj[0] == MessageEncoder.message_key:
                if obj[1]:
                    obj[3] = mark_safe(obj[3])
                return Message(*obj[2:])
            return [self.process_messages(item) for item in obj]
        if isinstance(obj, dict):
            return {key: self.process_messages(value) for key, value in obj.items()}
        return obj

    def decode(self, s, **kwargs):
        decoded = super().decode(s, **kwargs)
        return self.process_messages(decoded)


class MessagePartSerializer:
    def dumps(self, obj):
        return [
            json.dumps(
                o,
                separators=(",", ":"),
                cls=MessageEncoder,
            )
            for o in obj
        ]


class MessagePartGatherSerializer:
    def dumps(self, obj):
        return ("[" + ",".join(obj) + "]").encode("latin-1")


class MessageSerializer:
    def loads(self, data):
        return json.loads(data.decode("latin-1"), cls=MessageDecoder)


class CookieStorage(BaseStorage):
    cookie_name = "messages"
    max_cookie_size = 2048
    not_finished = "__messagesnotfinished__"
    not_finished_json = json.dumps("__messagesnotfinished__")
    key_salt = "aquilify.core.messages"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.signer = signing.get_cookie_signer(salt=self.key_salt)

    def _get(self, *args, **kwargs):
        data = self.request.cookies.get(self.cookie_name)
        messages = self._decode(data)
        all_retrieved = not (messages and messages[-1] == self.not_finished)
        if messages and not all_retrieved:
            messages.pop()
        return messages, all_retrieved

    def _update_cookie(self, encoded_data, response):
        if encoded_data:
            response._set_cookie(
                self.cookie_name,
                encoded_data,
                domain=settings.SESSION_COOKIE_DOMAIN,
                secure=settings.SESSION_COOKIE_SECURE or None,
                httponly=settings.SESSION_COOKIE_HTTPONLY or None,
                samesite=settings.SESSION_COOKIE_SAMESITE,
            )
        else:
            response._del_cookie(
                self.cookie_name
            )

    def _store(self, messages, response, remove_oldest=True, *args, **kwargs):
        unstored_messages = []
        serialized_messages = MessagePartSerializer().dumps(messages)
        encoded_data = self._encode_parts(serialized_messages)
        if self.max_cookie_size:
            cookie = SimpleCookie()

            def is_too_large_for_cookie(data):
                return data and len(cookie.value_encode(data)[1]) > self.max_cookie_size

            def compute_msg(some_serialized_msg):
                return self._encode_parts(
                    some_serialized_msg + [self.not_finished_json],
                    encode_empty=True,
                )

            if is_too_large_for_cookie(encoded_data):
                if remove_oldest:
                    idx = bisect_keep_right(
                        serialized_messages,
                        fn=lambda m: is_too_large_for_cookie(compute_msg(m)),
                    )
                    unstored_messages = messages[:idx]
                    encoded_data = compute_msg(serialized_messages[idx:])
                else:
                    idx = bisect_keep_left(
                        serialized_messages,
                        fn=lambda m: is_too_large_for_cookie(compute_msg(m)),
                    )
                    unstored_messages = messages[idx:]
                    encoded_data = compute_msg(serialized_messages[:idx])

        self._update_cookie(encoded_data, response)
        return unstored_messages

    def _encode_parts(self, messages, encode_empty=False):
        if messages or encode_empty:
            return self.signer.sign_object(
                messages, serializer=MessagePartGatherSerializer, compress=True
            )

    def _encode(self, messages, encode_empty=False):
        serialized_messages = MessagePartSerializer().dumps(messages)
        return self._encode_parts(serialized_messages, encode_empty=encode_empty)

    def _decode(self, data):
        if not data:
            return None
        try:
            return self.signer.unsign_object(data, serializer=MessageSerializer)
        except (signing.BadSignature, binascii.Error, json.JSONDecodeError):
            pass
        self.used = True
        return None


def bisect_keep_left(a, fn):
    lo = 0
    hi = len(a)
    while lo < hi:
        mid = (lo + hi) // 2
        if fn(a[: mid + 1]):
            hi = mid
        else:
            lo = mid + 1
    return lo


def bisect_keep_right(a, fn):
    lo = 0
    hi = len(a)
    while lo < hi:
        mid = (lo + hi) // 2
        if fn(a[mid:]):
            lo = mid + 1
        else:
            hi = mid
    return lo



from .column import Column
from typing import ClassVar, Optional

class PrimaryKeyField(Column):
    ALLOWED_PRIMARY_TYPES = {"INTEGER", "BIGINT", "UUID", "VARCHAR", "CHAR", "TEXT"}

    def __init__(
        self,
        type: str = "INTEGER",
        key_length: Optional[int] = None,
        autoincrement: bool = True,
        **kwargs
    ) -> None:
        self.validate_primary_key_type(type)
        self.validate_primary_key_length(type, key_length)
        
        super().__init__(
            type=self.construct_primary_key_type(type, key_length),
            primary_key=True,
            autoincrement=autoincrement,
            **kwargs
        )

    @staticmethod
    def validate_primary_key_type(primary_key_type: str) -> None:
        if primary_key_type.upper() not in PrimaryKeyField.ALLOWED_PRIMARY_TYPES:
            raise ValueError(
                f"Invalid primary key type: '{primary_key_type}'. Allowed types: {', '.join(PrimaryKeyField.ALLOWED_PRIMARY_TYPES)}"
            )

    @staticmethod
    def validate_primary_key_length(primary_key_type: str, primary_key_length: Optional[int]) -> None:
        if primary_key_length is not None and primary_key_type.upper() not in {"VARCHAR", "CHAR"}:
            raise ValueError(
                f"Primary key length is only applicable for 'VARCHAR' and 'CHAR' types, not for '{primary_key_type}'"
            )

    @staticmethod
    def construct_primary_key_type(primary_key_type: str, primary_key_length: Optional[int]) -> str:
        if primary_key_length is not None:
            return f"{primary_key_type.upper()}({primary_key_length})"
        return primary_key_type.upper()


class VarCharField(Column):
    DEFAULT_MAX_LENGTH: ClassVar[int] = 255
    MAX_ALLOWED_LENGTH: ClassVar[int] = 1000000

    def __init__(self, max_length: int = DEFAULT_MAX_LENGTH, **kwargs) -> None:
        self.max_length: int = self.validate_max_length(max_length)
        super().__init__(type=self.construct_type(self.max_length), **kwargs)

    @classmethod
    def construct_type(cls, max_length: int) -> str:
        return f"VARCHAR({max_length})"

    @staticmethod
    def validate_max_length(max_length: int) -> int:
        VarCharField._validate_positive_integer(max_length)
        VarCharField._validate_within_allowed_limit(max_length)
        return max_length

    @staticmethod
    def _validate_positive_integer(value: int) -> None:
        if not isinstance(value, int) or value <= 0:
            raise ValueError(f"Max length should be a positive integer. Received: {value}")

    @classmethod
    def _validate_within_allowed_limit(cls, value: int) -> None:
        if value > cls.MAX_ALLOWED_LENGTH:
            raise ValueError(
                f"Max length exceeds the allowed limit of {cls.MAX_ALLOWED_LENGTH}. Received: {value}"
            )
            
class CharField(Column):
    DEFAULT_MAX_LENGTH: ClassVar[int] = 255
    MAX_ALLOWED_LENGTH: ClassVar[int] = 1000000

    def __init__(self, max_length: int = DEFAULT_MAX_LENGTH, **kwargs) -> None:
        validated_length: int = self._validate_max_length(max_length)
        column_type: str = self._construct_column_type(validated_length)
        super().__init__(type=column_type, **kwargs)
        self.max_length: int = validated_length

    @classmethod
    def _construct_column_type(cls, max_length: int) -> str:
        return f"CHAR({max_length})"

    @staticmethod
    def _validate_max_length(max_length: int) -> int:
        CharField._validate_positive_integer(max_length)
        CharField._validate_within_allowed_limit(max_length)
        return max_length

    @staticmethod
    def _validate_positive_integer(value: int) -> None:
        if not isinstance(value, int) or value <= 0:
            raise ValueError(f"Max length should be a positive integer. Received: {value}")

    @classmethod
    def _validate_within_allowed_limit(cls, value: int) -> None:
        if value > cls.MAX_ALLOWED_LENGTH:
            raise ValueError(
                f"Max length exceeds the allowed limit of {cls.MAX_ALLOWED_LENGTH}. Received: {value}"
            )

class RealField(Column):
    DEFAULT_PRECISION: ClassVar[int] = 10
    DEFAULT_SCALE: ClassVar[int] = 5
    MAX_PRECISION: ClassVar[int] = 38
    MAX_SCALE: ClassVar[int] = 38

    def __init__(self, precision: int = DEFAULT_PRECISION, scale: int = DEFAULT_SCALE, **kwargs) -> None:
        self.validate_precision_scale(precision, scale)
        column_type = self.construct_column_type(precision, scale)
        super().__init__(type=column_type, **kwargs)
        self.precision = precision
        self.scale = scale

    @classmethod
    def construct_column_type(cls, precision: int, scale: int) -> str:
        return f"REAL({precision}, {scale})"

    @staticmethod
    def validate_precision_scale(precision: int, scale: int) -> None:
        if not isinstance(precision, int) or not isinstance(scale, int):
            raise TypeError("Precision and scale must be integers.")
        if precision <= 0 or scale < 0:
            raise ValueError("Precision must be a positive integer and scale must be a non-negative integer.")
        if precision > RealField.MAX_PRECISION or scale > RealField.MAX_SCALE:
            raise ValueError(f"Precision and scale should not exceed: Precision <= {RealField.MAX_PRECISION}, "
                             f"Scale <= {RealField.MAX_SCALE}")
        if precision < scale:
            raise ValueError("Precision must be greater than or equal to the scale.")

class TextField(Column):
    DEFAULT_MAX_LENGTH: ClassVar[int] = 65535
    MIN_LENGTH: ClassVar[int] = 1
    MAX_ALLOWED_LENGTH: ClassVar[int] = 1000000 

    def __init__(self, max_length: int = DEFAULT_MAX_LENGTH, **kwargs) -> None:
        self.validate_max_length(max_length)
        super().__init__(type=f"TEXT({max_length})", **kwargs)
        self.max_length = max_length

    @classmethod
    def validate_max_length(cls, max_length: int) -> None:
        cls._validate_positive_integer(max_length)
        if max_length < cls.MIN_LENGTH:
            raise ValueError(f"Max length must be at least {cls.MIN_LENGTH}.")
        if max_length > cls.MAX_ALLOWED_LENGTH:
            raise ValueError(f"Max length exceeds the allowed limit of {cls.MAX_ALLOWED_LENGTH}.")

    @staticmethod
    def _validate_positive_integer(value: int) -> None:
        if not isinstance(value, int) or value <= 0:
            raise ValueError("Max length should be a positive integer.")

class BigintField(Column):
    DEFAULT_MIN_VALUE: ClassVar[int] = -9223372036854775808  
    DEFAULT_MAX_VALUE: ClassVar[int] = 9223372036854775807   

    def __init__(self, min_value: int = DEFAULT_MIN_VALUE, max_value: int = DEFAULT_MAX_VALUE, **kwargs) -> None:
        self._validate_integer(min_value, "min_value")
        self._validate_integer(max_value, "max_value")
        self._validate_min_max_values(min_value, max_value)
        super().__init__(type="BIGINT", **kwargs)
        self.min_value = min_value
        self.max_value = max_value

    @staticmethod
    def _validate_integer(value: int, field_name: str) -> None:
        if not isinstance(value, int):
            raise TypeError(f"{field_name.capitalize()} should be an integer.")

    @staticmethod
    def _validate_min_max_values(min_value: int, max_value: int) -> None:
        if min_value >= max_value:
            raise ValueError("Minimum value should be less than maximum value.")
        
class BlobField(Column):
    DEFAULT_MAX_LENGTH: ClassVar[int] = 16777216 

    def __init__(self, max_length: int = DEFAULT_MAX_LENGTH, **kwargs) -> None:
        self.validate_max_length(max_length)
        super().__init__(type=f"BLOB({max_length})", **kwargs)
        self.max_length = max_length

    @classmethod
    def validate_max_length(cls, max_length: int) -> None:
        cls._validate_positive_integer(max_length)
        if max_length <= 0:
            raise ValueError("Max length should be greater than 0.")

    @staticmethod
    def _validate_positive_integer(value: int) -> None:
        if not isinstance(value, int) or value <= 0:
            raise ValueError("Max length should be a positive integer.")
        
class BooleanField(Column):
    def __init__(self, **kwargs) -> None:
        super().__init__(type="BOOLEAN", **kwargs)
        
class IntegerField(Column):
    ALLOWED_INTEGER_TYPES = {"INTEGER", "BIGINT", "SMALLINT"}

    def __init__(
        self,
        type: str = "INTEGER",
        **kwargs
    ) -> None:
        validated_type = self.validate_integer_type(type)
        
        super().__init__(
            type=validated_type,
            **kwargs
        )

    @classmethod
    def validate_integer_type(cls, integer_type: str) -> str:
        validated_type = integer_type.upper() if integer_type.upper() in cls.ALLOWED_INTEGER_TYPES else "INTEGER"
        
        return cls.get_valid_integer_type(validated_type)

    @staticmethod
    def get_valid_integer_type(validated_type: str) -> str:
        return validated_type if validated_type in IntegerField.ALLOWED_INTEGER_TYPES else IntegerField.default_integer_type()

    @staticmethod
    def default_integer_type() -> str:
        return "INTEGER"


import inspect

from typing import (
    Optional,
    List,
    Type,
    Callable,
    Awaitable,
    TypeVar,
    Tuple,
    Pattern,
    Dict,
    Any
)

T = TypeVar("T")

RouteInfo = Tuple[
    str, List[str], Callable[..., Awaitable[T]], List[str], Pattern, Type[T], str, Dict[str, Any]
]
Routes = List[RouteInfo]
WebSocketRoute = Tuple[str, Callable[..., Awaitable[T]]]
WebSocketRoutes = List[WebSocketRoute]

routes: Routes = []
websockets: WebSocketRoutes = []

class ImproperlyConfigured(Exception):
    pass

def rule(
    path: str,
    endpoint: Optional[Callable[..., Awaitable[Any]]] = None,
    **kwargs: Any
) -> Optional[List[Tuple]]:
    """
    Define an HTTP routing rule.

    Args:
    - path (str): The URL path pattern.
    - endpoint (Optional[Callable[..., Awaitable[T]]]): The function to be executed for the route.
    - **kwargs (Any): Additional keyword arguments.

    Supported **kwargs:
    - response_model (Optional[Type[T]]): The response model type.
    - strict_slashes (bool): Whether to consider trailing slashes.
    - methods (List[str]): HTTP methods supported by the route.
    - name (Optional[str]): The name of the route.

    Returns:
    - Optional[List[Tuple[str, EndpointType]]]: A list containing a tuple of path and endpoint.

    Example:
    >>> rule('/home', home_handler, methods=['GET', 'POST'], name='home_route')
    """
    if not path.startswith('/'):
        raise ImproperlyConfigured("Paths must start with '/'.")
    
    response_model = kwargs.pop('response_model', None)
    strict_slashes = kwargs.pop('strict_slashes', True)
    methods = kwargs.pop('methods', None)
        
    return (path, methods, endpoint, strict_slashes, response_model, endpoint, kwargs.get('name'))

def rule_all(
    path: str,
    endpoint: Callable[..., Awaitable[T]] = None,
    response_model: Optional[Type[T]] = None,
    strict_slashes: bool = True,
    name: Optional[str] = None
) -> Callable[..., Awaitable[T]]:
    """
    Define an HTTP routing rule for all methods.

    Args:
    - path (str): The URL path pattern.
    - endpoint (Callable[..., Awaitable[T]]): The function to be executed for the route.
    - response_model (Optional[Type[T]]): The response model type.
    - strict_slashes (bool): Whether to consider trailing slashes.
    - name (Optional[str]): The name of the route.

    Returns:
    - Callable[..., Awaitable[T]]: The function to handle the route.

    Example:
    >>> rule_all('/user', user_handler, UserModel, strict_slashes=False, name='user_route')
    """
    
    handler = endpoint
    allowed_methods = [
        "GET", "POST", "PUT", "DELETE",
        "HEAD", "OPTIONS", "PATCH", "TRACE"
    ]
    
    if not path.startswith('/'):
        raise ImproperlyConfigured("Paths must start with '/'.")
    
    return (
        path,
        allowed_methods,
        handler,
        strict_slashes,
        response_model,
        endpoint,
        name
    )
    
def re_rule(
    path_regex: str,
    endpoint: Optional[Callable[..., Awaitable[Any]]] = None,
    **kwargs: Any
    ) -> Optional[List[Tuple]]:
    """
    Define a regex-based HTTP routing rule.

    Args:
    - path_regex (str): The regex pattern for the URL path.
    - endpoint (Optional[Callable[..., Awaitable[T]]]): The function to be executed for the route.
    - **kwargs (Any): Additional keyword arguments.

    Supported **kwargs:
    - response_model (Optional[Type[T]]): The response model type.
    - strict_slashes (bool): Whether to consider trailing slashes.
    - methods (List[str]): HTTP methods supported by the route.
    - name (Optional[str]): The name of the route.

    Returns:
    - Optional[List[Tuple[str, EndpointType]]]: A list containing a tuple of path and endpoint.

    Example:
    >>> re_rule('/user/\\d+', user_handler, methods=['GET', 'PUT'], name='user_route')
    """
    if endpoint is None or not (inspect.iscoroutinefunction(endpoint) or inspect.isasyncgenfunction(endpoint)):
        raise ImproperlyConfigured("Invalid handler function provided for adding a route.")

    response_model = kwargs.pop('response_model', None)
    methods = kwargs.pop('methods', None)

    return (path_regex, methods, endpoint, kwargs.get('strict_slashes', False), response_model, endpoint, kwargs.get('name'))


def websocket(
    path: str,
    endpoint: Optional[Callable[..., Awaitable[T]]] = None,
) -> None:
    """
    Define a WebSocket routing rule.

    Args:
    - path (str): The URL path pattern.
    - endpoint (Optional[Callable[..., Awaitable[T]]]): The function to be executed for the route.
    
    Example:
    >>> websocket('/ws', ws_handler)
    """
    handler = endpoint
    if handler is None:
        raise ValueError("Handler function is required for adding a websocket route.")
    
    if not (inspect.iscoroutinefunction(handler) or inspect.isasyncgenfunction(handler)):
        raise TypeError("ASGI Websocket can only register asynchronous functions.")
        
    if not path.startswith('/'):
        raise TypeError("Websocket paths must start with '/'.")
        
    return (path, endpoint)


import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.image import MIMEImage
import ssl
import sys
import logging
from typing import List

from aquilify.settings import settings

class Mailer:
    def __init__(self, config_name=None):
        """
        Initialize the Mailer.

        Args:
            config_name (str, optional): The name of the email configuration.

        Usage:
            mailer = Mailer(config_name='config_name')
        """
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)

        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        try:
            self.configurations = settings.BASE_MAILER
            if config_name:
                self.config = next((config for config in self.configurations if config['NAME'] == config_name), None)
                if not self.config:
                    self.logger.error(f"Configuration '{config_name}' not found.")
                    sys.exit(1)
            else:
                self.config = self.configurations[0]
        except ImportError:
            self.logger.error("No settings.py found.")
            sys.exit(1)

    def _send_email(self, message, timeout=None):
        context = ssl.create_default_context() if self.config['SSL'] else None
        
        try:
            with smtplib.SMTP(self.config['SMTP'], self.config['PORT'], timeout=timeout) as server:
                if self.config['SSL']:
                    server.starttls(context=context)
                server.login(self.config['USERNAME'], self.config['PASSWORD'])
                server.sendmail(message['From'], message['To'], message.as_string())
            return True
        except Exception as e:
            self.logger.error("Error sending email:", exc_info=True)
            return False

    def send_email(self, to_email, subject, body=None, html_body=None, from_email=None, cc=None, bcc=None,
                   reply_to=None, attachments=None, inline_images=None, headers=None, charset='utf-8', timeout=None):
        """
        Send an email.

        Args:
            to_email (str): The recipient's email address.
            subject (str): The email subject.
            body (str, optional): The plain text email body.
            html_body (str, optional): The HTML email body.
            from_email (str, optional): The sender's email address.
            cc (str, optional): The CC email addresses.
            bcc (str, optional): The BCC email addresses.
            reply_to (str, optional): The reply-to email address.
            attachments (List[str], optional): List of file paths for email attachments.
            inline_images (List[str], optional): List of file paths for inline images.
            headers (dict, optional): Additional email headers.
            charset (str, optional): The character encoding.
            timeout (int, optional): Timeout for sending the email.

        Returns:
            bool: True if the email was sent successfully, False otherwise.

        Usage:
            success = mailer.send_email(
                to_email='recipient@example.com',
                subject='Test Email',
                body='This is a test email.',
                from_email='sender@example.com',
                cc='cc@example.com',
                bcc='bcc@example.com',
                reply_to='reply@example.com',
                attachments=['file1.txt', 'file2.pdf'],
                inline_images=['image1.png', 'image2.jpg'],
                headers={'X-Custom-Header': 'Custom Value'},
                charset='utf-8',
                timeout=30
            )
        """
        if not body and not html_body:
            self.logger.error("Both 'body' and 'html_body' cannot be None.")
            return

        if not from_email:
            from_email = self.config['DEFAULT_SENDER']

        msg = MIMEMultipart()
        msg['From'] = from_email
        msg['To'] = to_email
        msg['Subject'] = subject

        if cc:
            msg['Cc'] = cc
        if bcc:
            msg['Bcc'] = bcc
        if reply_to:
            msg.add_header('reply-to', reply_to)

        if headers:
            for key, value in headers.items():
                msg.add_header(key, value)

        if html_body:
            msg.attach(MIMEText(html_body, 'html', charset))
        elif body:
            msg.attach(MIMEText(body, 'plain', charset))

        if attachments:
            for attachment in attachments:
                with open(attachment, 'rb') as file:
                    part = MIMEApplication(file.read(), Name=attachment)
                    part['Content-Disposition'] = f'attachment; filename="{attachment}"'
                    msg.attach(part)

        if inline_images:
            for inline_image in inline_images:
                with open(inline_image, 'rb') as file:
                    img = MIMEImage(file.read(), name=inline_image)
                    img.add_header('Content-ID', f'<{inline_image}>')
                    img.add_header('Content-Disposition', 'inline', filename=inline_image)
                    msg.attach(img)

        return self._send_email(msg, timeout=timeout)

    def send_bulk_email(self, to_emails: List[str], subject, body=None, html_body=None, from_email=None,
                        cc=None, bcc=None, reply_to=None, attachments=None, inline_images=None,
                        headers=None, charset='utf-8', timeout=None):
        """
        Send bulk email to multiple recipients.

        Args:
            to_emails (List[str]): List of recipient email addresses.
            subject (str): The email subject.
            body (str, optional): The plain text email body.
            html_body (str, optional): The HTML email body.
            from_email (str, optional): The sender's email address.
            cc (str, optional): The CC email addresses.
            bcc (str, optional): The BCC email addresses.
            reply_to (str, optional): The reply-to email address.
            attachments (List[str], optional): List of file paths for email attachments.
            inline_images (List[str], optional): List of file paths for inline images.
            headers (dict, optional): Additional email headers.
            charset (str, optional): The character encoding.
            timeout (int, optional): Timeout for sending the email.

        Returns:
            int: The number of successful email sends.

        Usage:
            success_count = mailer.send_bulk_email(
                to_emails=['recipient1@example.com', 'recipient2@example.com'],
                subject='Bulk Test Email',
                body='This is a bulk test email.',
                from_email='sender@example.com',
                attachments=['file1.txt', 'file2.pdf'],
                charset='utf-8',
                timeout=30
            )
        """
        success_count = 0
        for to_email in to_emails:
            if self.send_email(to_email, subject, body, html_body, from_email, cc, bcc, reply_to,
                               attachments, inline_images, headers, charset, timeout):
                success_count += 1
        return success_count

    def send_email_timeout(self, to_email, subject, body=None, html_body=None, from_email=None, cc=None, bcc=None,
                          reply_to=None, attachments=None, inline_images=None, headers=None, charset='utf-8',
                          timeout=30):
        """
        Send an email with a specified timeout.

        Args:
            to_email (str): The recipient's email address.
            subject (str): The email subject.
            body (str, optional): The plain text email body.
            html_body (str, optional): The HTML email body.
            from_email (str, optional): The sender's email address.
            cc (str, optional): The CC email addresses.
            bcc (str, optional): The BCC email addresses.
            reply_to (str, optional): The reply-to email address.
            attachments (List[str], optional): List of file paths for email attachments.
            inline_images (List[str], optional): List of file paths for inline images.
            headers (dict, optional): Additional email headers.
            charset (str, optional): The character encoding.
            timeout (int, optional): Timeout for sending the email.

        Returns:
            bool: True if the email was sent successfully, False otherwise.

        Usage:
            success = mailer.send_email_timeout(
                to_email='recipient@example.com',
                subject='Test Email',
                body='This is a test email.',
                from_email='sender@example.com',
                attachments=['file1.txt', 'file2.pdf'],
                charset='utf-8',
                timeout=30
            )
        """
        return self.send_email(to_email, subject, body, html_body, from_email, cc, bcc, reply_to, attachments,
                               inline_images, headers, charset, timeout)

    def send_template_email(self, to_email, subject, template_path, context=None, from_email=None, cc=None,
                            bcc=None, reply_to=None, attachments=None, inline_images=None, headers=None,
                            charset='utf-8', timeout=None):
        """
        Send an email using an HTML template.

        Args:
            to_email (str): The recipient's email address.
            subject (str): The email subject.
            template_path (str): The path to the HTML email template.
            context (dict, optional): The context for template rendering.
            from_email (str, optional): The sender's email address.
            cc (str, optional): The CC email addresses.
            bcc (str, optional): The BCC email addresses.
            reply_to (str, optional): The reply-to email address.
            attachments (List[str], optional): List of file paths for email attachments.
            inline_images (List[str], optional): List of file paths for inline images.
            headers (dict, optional): Additional email headers.
            charset (str, optional): The character encoding.
            timeout (int, optional): Timeout for sending the email.

        Returns:
            bool: True if the email was sent successfully, False otherwise.

        Usage:
            success = mailer.send_template_email(
                to_email='recipient@example.com',
                subject='Template Email',
                template_path='template.html',
                context={'name': 'John Doe'},
                from_email='sender@example.com',
                attachments=['file1.txt', 'file2.pdf'],
                charset='utf-8',
                timeout=30
            )
        """
        if not context:
            context = {}

        with open(template_path, 'r') as template_file:
            template = template_file.read()

        html_body = template.format(**context)
        return self.send_email(to_email, subject, html_body=html_body, from_email=from_email, cc=cc, bcc=bcc,
                               reply_to=reply_to, attachments=attachments, inline_images=inline_images,
                               headers=headers, charset=charset, timeout=timeout)



CONFIG_CFG = """[Aquilify]
path = /home/pawan/PythonProjects/testproject/myapp
project_name = myapp
version = 1.13
compiler_path = /home/pawan/PythonProjects/testproject/myapp/.aquilify
settings = /home/pawan/PythonProjects/testproject/myapp/settings.py

[ASGI_SERVER]
server = NETIX
host = 0.0.0.0
port = 8000
debug = True
reload = False
instance = asgi:app
"""


import http
import typing
import warnings

__all__ = ("HTTPException", "WebSocketException")


class HTTPException(Exception):
    def __init__(
        self,
        status_code: int,
        detail: typing.Optional[str] = None,
        headers: typing.Optional[dict] = None,
    ) -> None:
        if detail is None:
            detail = http.HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.detail = detail
        self.headers = headers

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"



class MagicFilter(object):
    def __init__(self, query: str, variables: tuple = None, parameters: dict[str, object] = None) -> None:

        """
        Magic filter used for advanced processing of sql queries

        :param query: Initial query
        :param variables: Query variables
        :param parameters: Parameters saved after writing the query to a new expression
        """

        self.query = query

        self.variables = variables or tuple()
        self.parameters = parameters or dict()

    @staticmethod
    def _serialize_variable(variable: object) -> tuple:
        if isinstance(variable, MagicFilter):
            return variable.variables
        else:
            return (variable,)

    @staticmethod
    def _serialize_query(query: object) -> str:
        if isinstance(query, MagicFilter):
            return query.query
        else:
            return "?"

    @staticmethod
    def _wrap_query(query: str, wrap: bool) -> str:
        return f"({query})" if wrap else query

    def _get_expression(self, other: object, operator: str, *, wrap: bool = False) -> "MagicFilter":
        return MagicFilter(
            f"{self._wrap_query(self.query, wrap)} "
            f"{operator} "
            f"{self._wrap_query(self._serialize_query(other), wrap)}",

            (*self.variables, *self._serialize_variable(other)), self.parameters
        )

    def _invert_expression(self) -> "MagicFilter":
        return MagicFilter(
            f"NOT ({self.query})", self.variables, self.parameters
        )

    def __eq__(self, other: object) -> "MagicFilter":
        return self._get_expression(other, "=")

    def __ne__(self, other: object) -> "MagicFilter":
        return self._get_expression(other, "!=")

    def __ge__(self, other: object) -> "MagicFilter":
        return self._get_expression(other, ">=")

    def __gt__(self, other: object) -> "MagicFilter":
        return self._get_expression(other, ">")

    def __le__(self, other: object) -> "MagicFilter":
        return self._get_expression(other, "<=")

    def __lt__(self, other: object) -> "MagicFilter":
        return self._get_expression(other, "<")

    def __and__(self, other: object) -> "MagicFilter":
        return self._get_expression(other, "AND", wrap=True)

    def __or__(self, other: object) -> "MagicFilter":
        return self._get_expression(other, "OR", wrap=True)

    def __invert__(self) -> "MagicFilter":
        return self._invert_expression()



import secrets
import base64

from typing import Dict, Any, Optional
from datetime import datetime, timedelta
from aquilify.core import signing

from aquilify.wrappers import Request, Response
from aquilify.settings.sessions import SessionConfigSettings
from aquilify.utils.module_loading import import_string
from aquilify.settings import settings

_settings = SessionConfigSettings()

class BeforeSessionStage:
    def __init__(
        self
    ) -> None:
        self.sessions: Dict[str] = {}
        self.serializer = signing
        self.session_lifetime: timedelta = timedelta(minutes=_settings.fetch().get('session_lifetime'))
        self.cookie_name: str = _settings.fetch().get('cookie_name')

    async def __call__(self, request: Request) -> Any:
        session_id = self._get_session_id(request.cookies.get(self.cookie_name))
        if session_id is None:
            session_id = self._generate_session_id()
        storage = None
        
        for item in settings.STORAGE_BACKEND:
            if "sessions" in item and "cookie" in item["sessions"]:
                storage = item["sessions"]["cookie"]
                break
        if not storage:
            raise ValueError("Either storage SESSION_BACKEND isn't found! or may not been configured properly!")
        func = import_string(storage)
        request.scope['session'] = self.sessions.setdefault(session_id, func(session_id))
        self._cleanup_sessions()
        await self._regenerate_expired_session(request)
        return request

    def _generate_session_id(self) -> str:
        return base64.urlsafe_b64encode(secrets.token_bytes(64)).decode('utf-8')

    def _get_session_id(self, signed_session_id: Optional[str]) -> Optional[str]:
        if signed_session_id:
            try:
                return self.serializer.loads(signed_session_id, settings.SECRET_KEY, max_age=_settings.fetch().get('max_age'))
            except signing.BadSignature:
                pass
        return None

    def _cleanup_sessions(self) -> None:
        now = datetime.now()
        expired_sessions = [
            sid for sid, session in self.sessions.items()
            if now - session._created_at > self.session_lifetime
        ]
        for sid in expired_sessions:
            self.sessions.pop(sid)

    async def _regenerate_expired_session(self, request: Request) -> None:
        if 'expired_session' in request.cookies:
            signed_session_id = request.cookies.get('expired_session')
            session_id = self._get_session_id(signed_session_id)
            if session_id in self.sessions:
                old_session = self.sessions.pop(session_id)
                self.sessions[old_session._session_id] = old_session

    def update_cookie_name(self, cookie_name: str) -> None:
        self.cookie_name = cookie_name

class AfterSessionStage:
    def __init__(
        self
    ) -> None:
        self.serializer = signing
        self.max_age: int = _settings.fetch().get('max_age')
        self.secure: bool = _settings.fetch().get('secure')
        self.httponly: bool = _settings.fetch().get('httponly')
        self.samesite: str = _settings.fetch().get('samesite')
        self.cookie_name: str = _settings.fetch().get('cookie_name')
        self.domain: str = getattr(settings, 'SESSION_COOKIE_DOMAIN', None)

    async def __call__(self, request: Request, response: Response) -> Any:
        session_id = request.scope['session']._session_id if 'session' in request.scope else None
        if session_id:
            await self._set_cookie(response, session_id)
            await self._regenerate_expired_session(request, response)
        return response

    async def _set_cookie(self, response: Response, session_id: str) -> None:
        signed_session_id = self.serializer.dumps(session_id, settings.SECRET_KEY)
        await response.set_cookie(
            self.cookie_name,
            signed_session_id,
            max_age=self.max_age,
            secure=self.secure,
            httponly=self.httponly,
            samesite=self.samesite,
            domain=self.domain
        )

    async def _regenerate_expired_session(self, request: Request, response: Response) -> None:
        if 'expired_session' in request.cookies:
            signed_session_id = request.cookies.get('expired_session')
            session_id = self._get_session_id(signed_session_id)
            if session_id in self.sessions:
                old_session = self.sessions.pop(session_id)
                self.sessions[old_session._session_id] = old_session
                await response.set_cookie('expired_session', '', max_age=0)
                await self._set_cookie(response, old_session._session_id)




from typing import Any, Union, List, Optional

from . import constants
from .storage import default_storage

__all__ = (
    "add_message",
    "get_messages",
    "get_level",
    "set_level",
    "debug",
    "info",
    "success",
    "warning",
    "error",
    "MessageFailure",
)

class MessageFailure(Exception):
    pass

def add_message(request: Any, level: int, message: str, extra_tags: str = "", fail_silently: bool = False) -> None:
    """
    Add a message to the request.

    Args:
        request (Any): Request object.
        level (int): Message level (DEBUG, INFO, SUCCESS, WARNING, ERROR).
        message (str): Message content.
        extra_tags (str, optional): Extra tags for the message.
        fail_silently (bool, optional): If True, failures will not raise an error.

    Raises:
        TypeError: If the argument passed is not a valid request object.
        MessageFailure: If MessageMiddleware is not installed and fail_silently is False.
    """
    try:
        messages = request._messages
    except AttributeError:
        if not hasattr(request, "headers"):
            raise TypeError(
                f"add_message() argument must be a Request object, not '{request.__class__.__name__}'."
            )
        if not fail_silently:
            raise MessageFailure(
                "You cannot add messages without installing "
                "aquilify.core.messages.middleware.MessageMiddleware"
            )
    else:
        messages.add(level, message, extra_tags)

def get_messages(request: Any) -> List[Any]:
    """
    Get messages from the request.

    Args:
        request (Any): Request object.

    Returns:
        List[Any]: List of messages or an empty list if no messages found.
    """
    return getattr(request, "_messages", [])

def get_level(request: Any) -> int:
    """
    Get the current message level from the request.

    Args:
        request (Any): Request object.

    Returns:
        int: Current message level.
    """
    storage = getattr(request, "_messages", default_storage(request))
    return storage.level

def set_level(request: Any, level: int) -> bool:
    """
    Set the message level for the request.

    Args:
        request (Any): Request object.
        level (int): Message level to set.

    Returns:
        bool: True if the message level is set successfully, False otherwise.
    """
    if not hasattr(request, "_messages"):
        return False
    request._messages.level = level
    return True

def debug(request: Any, message: str, extra_tags: str = "", fail_silently: bool = False) -> None:
    """
    Add a debug message to the request.

    Args:
        request (Any): Request object.
        message (str): Debug message content.
        extra_tags (str, optional): Extra tags for the message.
        fail_silently (bool, optional): If True, failures will not raise an error.
    """
    add_message(request, constants.DEBUG, message, extra_tags=extra_tags, fail_silently=fail_silently)

def info(request: Any, message: str, extra_tags: str = "", fail_silently: bool = False) -> None:
    """
    Add an info message to the request.

    Args:
        request (Any): Request object.
        message (str): Info message content.
        extra_tags (str, optional): Extra tags for the message.
        fail_silently (bool, optional): If True, failures will not raise an error.
    """
    add_message(request, constants.INFO, message, extra_tags=extra_tags, fail_silently=fail_silently)

def success(request: Any, message: str, extra_tags: str = "", fail_silently: bool = False) -> None:
    """
    Add a success message to the request.

    Args:
        request (Any): Request object.
        message (str): Success message content.
        extra_tags (str, optional): Extra tags for the message.
        fail_silently (bool, optional): If True, failures will not raise an error.
    """
    add_message(request, constants.SUCCESS, message, extra_tags=extra_tags, fail_silently=fail_silently)

def warning(request: Any, message: str, extra_tags: str = "", fail_silently: bool = False) -> None:
    """
    Add a warning message to the request.

    Args:
        request (Any): Request object.
        message (str): Warning message content.
        extra_tags (str, optional): Extra tags for the message.
        fail_silently (bool, optional): If True, failures will not raise an error.
    """
    add_message(request, constants.WARNING, message, extra_tags=extra_tags, fail_silently=fail_silently)

def error(request: Any, message: str, extra_tags: str = "", fail_silently: bool = False) -> None:
    """
    Add an error message to the request.

    Args:
        request (Any): Request object.
        message (str): Error message content.
        extra_tags (str, optional): Extra tags for the message.
        fail_silently (bool, optional): If True, failures will not raise an error.
    """
    add_message(request, constants.ERROR, message, extra_tags=extra_tags, fail_silently=fail_silently)



import asyncio
from functools import lru_cache
from markupsafe import Markup
from html import escape

try:
    from jinja2 import Environment, FileSystemLoader, select_autoescape, Template, TemplateNotFound, TemplateError
except ImportError:
    Environment, FileSystemLoader, select_autoescape, Template, TemplateNotFound, TemplateError = None, None, None, None, None, None

from typing import (
    Any,
    Dict,
    Optional,
    Callable,
    List
)

from ..wrappers import Response, Request
from ..exception.__handler import handle_exception

from ..settings.templates import TemplateConfigSettings

try:
    _config_settings = TemplateConfigSettings()
    _config_settings.fetch()
    _settings = _config_settings.template_data[0]
except Exception as e:
    raise Exception("Template settings not found in settings.py, configure it before you use.")


class TemplateResponse:
    def __init__(
        self
    ):
        self._check_jinja2_library()
        
        self.template_paths = _settings.get('dirs') or ["templates"]
        self.default_context = None or {}
        self.autoescape = _settings['options'].get('autoscape') or True
        self.template_engine = _settings.get('backend').lower()
        self.cache_size = _settings['options'].get('cache_size') or True
        self.context_processors = _settings['options'].get('context_processors') or []
        self.flash_config = {'with_category': False, 'category_filter': ()}
        self.custom_filters = _settings['options'].get('filters') or {}
        self.custom_globals = _settings['options'].get('globals') or {}
        self.enable_template_cache = _settings['options'].get('enable_template_cache') or True
        self.custom_extensions = _settings['options'].get('extensions') or []
        self.csrf = _settings.get('csrf') or None

        if self.template_engine not in ["jinja2"]:
            raise ValueError("Unsupported template engine. Currently, only 'jinja2' is supported.")

        self.env: Environment = self._create_environment()

    def _check_jinja2_library(self):
        if Environment is None:
            raise ImportError("Jinja2 library is not installed. Please install it using 'pip install jinja2' or 'pip install aquilify[jinja2]'.")

    def _create_environment(self) -> Environment:
        if self.template_engine == "jinja2":
            loader = FileSystemLoader(self.template_paths)
            environment = Environment(
                loader=loader,
                autoescape=select_autoescape(['html', 'xml']) if self.autoescape else False,
                cache_size=self.cache_size if self.enable_template_cache else 0,
                extensions=self.custom_extensions,
            )
            environment.filters.update(self.custom_filters)
            environment.globals.update(self.custom_globals)
            return environment

    @lru_cache(maxsize=None)
    def _get_template(self, template_name: str) -> Template:
        try:
            return self.env.get_template(template_name)
        except TemplateNotFound as e:
            error_message = f"Template not found: {e.name}"
            raise FileNotFoundError(error_message) from e
        except TemplateError as e:
            error_message = f"Error loading template '{template_name}': {str(e)}"
            raise FileNotFoundError(error_message) from e

    def _inject_default_context(self, context: Dict[str, Any]) -> Dict[str, Any]:
        injected_context = self.default_context.copy()
        injected_context.update(context)
        return injected_context

    async def _add_url_generation(self, request: Request, context: Dict[str, Any], csrf_protect) -> Dict[str, Any]:
        if self.csrf is not None:
            context['flashes'] = await self._get_flashes(request)
            context['csrf_protect'] = csrf_protect
            return context
        context['flashes'] = await self._get_flashes(request)
        return context

    def _get_processor_name(self, processor) -> str:
        if hasattr(processor, '__call__'):
            if isinstance(processor, type):
                return f"Callable class '{processor.__name__}'"
            else:
                return f"Instance of callable class '{processor.__class__.__name__}'"
        else:
            return f"Function '{processor.__name__}'"

    def _run_context_processors(self, context: Dict[str, Any], request: Request) -> Dict[str, Any]:
        for processor in self.context_processors:
            processor_name = self._get_processor_name(processor)

            processed_context = processor(context, request)
            if not isinstance(processed_context, dict):
                raise ValueError(f"{processor_name} must return a dictionary.")
            context = processed_context
        return context
    
    async def _clear_flashes(self, request: Request):
        request.context['flash'] = {}

    async def flash(self, request: Request, message, category='message'):
        context = request.context.setdefault('flash', {})
        context.setdefault(category, []).append(message)

    async def _get_flashes(self, request: Request) -> Dict[str, List[str]]:
        with_category = self.flash_config.get('with_category', False)
        category_filter = self.flash_config.get('category_filter', ())

        flashes = request.context.pop('flash', {})

        if not with_category:
            return flashes

        if category_filter:
            filtered_flashes = {category: messages for category, messages in flashes.items() if category in category_filter}
            return filtered_flashes
        else:
            return flashes
            
    async def render(
        self,
        request: Request,
        template_name: str,
        context: Dict[str, Any] = None,
        status_code: int = 200,
        headers: Dict[str, str] = None,
        inherit: Optional[str] = None,
    ) -> Response:
        
        if context is None:
            context = {}
        
        token = None
        if self.csrf is not None:
            token = await self.csrf.generate_csrf_token(request.remote_addr)
            token_name = self.csrf.csrf_token_key
            csrf_protect = lambda: Markup(f'<input name="{token_name}" type="hidden" value="{escape(token)}"></input>') if token else ''

        csrf_protect = None
        
        template = await asyncio.to_thread(self._get_template, template_name)
        context = self._inject_default_context(context)
        context = await self._add_url_generation(request, context, csrf_protect)
        context = self._run_context_processors(context, request)

        try:
            if inherit:
                inherited_template = await asyncio.to_thread(self._get_template, inherit)
                content = await asyncio.to_thread(template.render, content=inherited_template.render(**context), **context)
            else:
                content = await asyncio.to_thread(template.render, **context)
        except TemplateNotFound as e:
            error_message = f"Template not found: {e.name}"
            raise FileNotFoundError(error_message) from e
        except TemplateError as e:
            await handle_exception(e)

        response = Response(
            content,
            content_type='text/html',
            status_code=status_code,
            headers=headers
        )
        if self.csrf is not None:
            await self.csrf.inject_csrf_token(response, token)
            response.headers['X-CSRF-TOKEN'] = token
        return response

    async def __call__(
        self,
        request: Request,
        template_name: str,
        context: Dict[str, Any] = None,
        status_code: int = 200,
        headers: Dict[str, str] = None,
        inherit: Optional[str] = None,
    ) -> Response:
        
        if context is None:
            context = {}
        
        token = None
        if self.csrf is not None:
            token = await self.csrf.generate_csrf_token(request.remote_addr)
            token_name = self.csrf.csrf_token_key
            csrf_protect = lambda: Markup(f'<input name="{token_name}" type="hidden" value="{escape(token)}"></input>') if token else ''

        else:
            csrf_protect = None
        
        template = await asyncio.to_thread(self._get_template, template_name)
        context = self._inject_default_context(context)
        context = await self._add_url_generation(request, context, csrf_protect)
        context = self._run_context_processors(context, request)

        try:
            if inherit:
                inherited_template = await asyncio.to_thread(self._get_template, inherit)
                content = await asyncio.to_thread(template.render, content=inherited_template.render(**context), **context)
            else:
                content = await asyncio.to_thread(template.render, **context)
        except TemplateNotFound as e:
            error_message = f"Template not found: {e.name}"
            raise FileNotFoundError(error_message) from e
        except TemplateError as e:
            await handle_exception(e)

        response = Response(
            content,
            content_type='text/html',
            status_code=status_code,
            headers=headers
        )
        if self.csrf is not None:
            await self.csrf.inject_csrf_token(response, token)
            response.headers['X-CSRF-TOKEN'] = token
        return response


class Colorib:
    # ANSI color codes for text colors
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    RESET = "\033[0m"  # Reset all text formatting

    # ANSI color codes for background colors
    BG_BLACK = "\033[40m"
    BG_RED = "\033[41m"
    BG_GREEN = "\033[42m"
    BG_YELLOW = "\033[43m"
    BG_BLUE = "\033[44m"
    BG_MAGENTA = "\033[45m"
    BG_CYAN = "\033[46m"
    BG_WHITE = "\033[47m"

    # Text style modifiers
    BOLD = "\033[1m"
    ITALIC = "\033[3m"
    UNDERLINE = "\033[4m"
    BLINK = "\033[5m"
    REVERSE = "\033[7m"
    CONCEAL = "\033[8m"

    # Colorama-like functions
    init = lambda: None  # Placeholder for Colorama's init function

    @classmethod
    def colorize(cls, text, color):
        return f"{color}{text}{cls.RESET}"

    @classmethod
    def style(cls, text, style):
        return f"{style}{text}{cls.RESET}"

    @classmethod
    def colorize_with_background(cls, text, color, bg_color):
        return f"{color}{bg_color}{text}{cls.RESET}"

    @classmethod
    def style_with_background(cls, text, style, bg_color):
        return f"{style}{bg_color}{text}{cls.RESET}"

    @classmethod
    def auto_reset(cls, text):
        return f"{text}{cls.RESET}"

    @classmethod
    def strip_ansi(cls, text):
        import re
        ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')
        return ansi_escape.sub('', text)

    @classmethod
    def print_color(cls, text, color):
        print(f"{color}{text}{cls.RESET}")

    @classmethod
    def disable(cls):
        cls.RESET = ''
        cls.BLACK = ''
        cls.RED = ''
        cls.GREEN = ''
        cls.YELLOW = ''
        cls.BLUE = ''
        cls.MAGENTA = ''
        cls.CYAN = ''
        cls.WHITE = ''
        cls.BG_BLACK = ''
        cls.BG_RED = ''
        cls.BG_GREEN = ''
        cls.BG_YELLOW = ''
        cls.BG_BLUE = ''
        cls.BG_MAGENTA = ''
        cls.BG_CYAN = ''
        cls.BG_WHITE = ''
        cls.BOLD = ''
        cls.ITALIC = ''
        cls.UNDERLINE = ''
        cls.BLINK = ''
        cls.REVERSE = ''
        cls.CONCEAL = ''

    @classmethod
    def enable(cls):
        cls.RESET = "\033[0m"
        cls.BLACK = "\033[30m"
        cls.RED = "\033[31m"
        cls.GREEN = "\033[32m"
        cls.YELLOW = "\033[33m"
        cls.BLUE = "\033[34m"
        cls.MAGENTA = "\033[35m"
        cls.CYAN = "\033[36m"
        cls.WHITE = "\033[37m"
        cls.BG_BLACK = "\033[40m"
        cls.BG_RED = "\033[41m"
        cls.BG_GREEN = "\033[42m"
        cls.BG_YELLOW = "\033[43m"
        cls.BG_BLUE = "\033[44m"
        cls.BG_MAGENTA = "\033[45m"
        cls.BG_CYAN = "\033[46m"
        cls.BG_WHITE = "\033[47m"
        cls.BOLD = "\033[1m"
        cls.ITALIC = "\033[3m"
        cls.UNDERLINE = "\033[4m"
        cls.BLINK = "\033[5m"
        cls.REVERSE = "\033[7m"
        cls.CONCEAL = "\033[8m"

    @classmethod
    def prompt(cls, text, color=WHITE):
        return input(f"{color}{text}{cls.RESET}")

    @classmethod
    def colorize_multiple(cls, text_parts, colors):
        if len(text_parts) != len(colors):
            raise ValueError("The number of text parts must match the number of colors")
        colorized_parts = [f"{color}{text}{cls.RESET}" for text, color in zip(text_parts, colors)]
        return "".join(colorized_parts)

    @classmethod
    def print_table(cls, data, headers=None, cell_width=12):
        if headers:
            cls.print_table_row(headers, header=True)
        for row in data:
            cls.print_table_row(row, cell_width=cell_width)

    @classmethod
    def print_table_row(cls, row, header=False, cell_width=12):
        format_str = f"{cls.BG_WHITE if header else ''}{{:<{cell_width}}}{cls.RESET}"
        formatted_row = [format_str.format(str(item)) for item in row]
        print("".join(formatted_row))



import importlib.util

class XSRFConfigSettings:
    def __init__(self, settings_module_path: str = "./settings.py") -> None:
        self.settings_module_path = settings_module_path
        self.template_data = []

    def fetch(self) -> None:
        try:
            with open(self.settings_module_path) as file:
                settings = self._load_settings_module()
                templates = getattr(settings, 'TEMPLATES', [])
                csrf = getattr(settings, 'CSRF', None)  # Fetch CSRF setting
                self._extract_templates(templates, csrf)

        except (FileNotFoundError, AttributeError, ValueError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings

    def _extract_templates(self, templates: list, csrf_setting) -> None:
        for template in templates:
            extracted_data = self._extract_template_data(template, csrf_setting)
            self.template_data.append(extracted_data)

    def _extract_template_data(self, template: dict, csrf_setting) -> dict:
        csrf = self._load_csrf(template.get('CSRF', None))

        if csrf_setting is None and csrf is not None:
            raise ValueError("CSRF setting cannot be used before setting it up in settings.py")

        return {
            'csrf': csrf
        }

    def _load_processor(self, processor_path: str):
        if processor_path:
            try:
                parts = processor_path.split('.')
                module_path = '.'.join(parts[:-1])
                callback_name = parts[-1]
                callback_module = importlib.import_module(module_path)
                callback_obj = getattr(callback_module, callback_name)
                if isinstance(callback_obj, type):
                    callback_obj = callback_obj()
                return callback_obj
            except (AttributeError, ImportError, ValueError) as e:
                print(f"Error fetching callback: {e}")
        return None

    def _load_csrf(self, csrf_string: str):
        if csrf_string:
            return self._load_processor(csrf_string)
        return csrf_string



import re
import inspect
import importlib
from importlib import import_module

from typing import (
    Any, 
    Awaitable, 
    Callable, 
    Dict, 
    List, 
    Optional, 
    Pattern, 
    Tuple, 
    Type, 
    TypeVar
)

from ..types import ASGIApp
from .__globals import Converter
from .schematic import Schematic
from ..exception.base import ImproperlyConfigured

from ..types import (
    Routes,
    WebSocketRoutes
)

T = TypeVar("T")

class ColoursCode:
    BG_YELLOW = "\033[35m"
    BLUE = "\033[34m"
    GREEN = "\033[32m"
    RED = "\033[31m"
    YELLOW = "\033[33m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    BOLD = "\033[1m"
    RESET = "\033[0m"

_routes: Routes = []
_links: Routes = []
_websockets: WebSocketRoutes = []
_handlers_set = set()

class DuplicateHandler:
    @staticmethod
    def _duplicate_handler(handler: Callable[..., Awaitable[T]]) -> None:
        if handler in _handlers_set:
            raise ImproperlyConfigured("Duplicate handler found in routes.")
        _handlers_set.add(handler)
    
    @staticmethod
    def _is_duplicate_route(path: str, endpoint: Callable[..., Awaitable[Any]]) -> bool:
        for route in _routes:
            if route[0] == path and route[2] == endpoint:
                return True
        return False
    
    @staticmethod
    def _is_duplicate_websocket(path: str, endpoint: Callable[..., Awaitable[Any]]) -> bool:
        for route in _websockets:
            if route[0] == path and route[2] == endpoint:
                return True
        return False
    
    @staticmethod
    def _is_duplicate_route_with_regex(path_regex: str, endpoint: Callable[..., Awaitable[Any]]) -> bool:
        return any(route[3].pattern == path_regex and route[2] == endpoint for route in _routes)
    
class RoutingHelpers:
    @staticmethod
    def _helper_inc_path(path: str, func_include: Callable[..., Awaitable[T]]):
        for func in func_include:
            if not len(func) == 7:
                pass
            else:
                sub_path, sub_methods, sub_handler, sub_strict_slashes, sub_response_model, sub_endpoint, sub_name = func
                
                if sub_handler is None or not (inspect.iscoroutinefunction(sub_handler) or inspect.isasyncgenfunction(sub_handler)):
                    raise ImproperlyConfigured("Invalid handler function provided for adding a route.")
                
                allowed_methods = {"GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS", "PATCH", "TRACE"}
                if sub_methods and not all(method.upper() in allowed_methods for method in map(str.upper, sub_methods)):
                    invalid_methods = [method for method in sub_methods if method.upper() not in allowed_methods]
                    raise ImproperlyConfigured(f"Invalid HTTP method(s) provided: {', '.join(invalid_methods)}")

                sub_methods = sub_methods or ["GET"]
                
                converted_path, path_regex = Converter()._regex_converter(path + sub_path, sub_strict_slashes, '')
                
                if DuplicateHandler._is_duplicate_route(converted_path, sub_handler):
                    raise ImproperlyConfigured("Duplicate endpoint detected for the same route.")
                
                _routes.append((
                    converted_path,
                    sub_methods,
                    sub_handler,
                    path_regex,
                    sub_response_model,
                    sub_endpoint
                ))
                _links.append((
                    path,
                    sub_endpoint,
                    sub_name
                ))
        return None
    
    @staticmethod
    def _helper_inc_websockets(path: str, func_include: Callable[..., Awaitable[T]]):
        for func in func_include:
            if not len(func) == 2:
                pass
            else:
                sub_path, sub_handler = func
                
                if sub_handler is None or not (inspect.iscoroutinefunction(sub_handler) or inspect.isasyncgenfunction(sub_handler)):
                    raise ImproperlyConfigured("Invalid handler function provided for adding a websocket route.")
                
                converted_path, path_regex = Converter()._regex_converter(path + sub_path, False)
                
                if DuplicateHandler._is_duplicate_websocket(converted_path, sub_handler):
                    raise ImproperlyConfigured("Duplicate endpoint detected for the same websocket route.")
                
                _websockets.append((
                    converted_path,
                    path_regex,
                    sub_handler
                ))
            
class HTTPRouting:
    @staticmethod
    def rule(
        path: str,
        endpoint: Optional[Callable[..., Awaitable[Any]]] = None,
        **kwargs: Any
    ) -> Optional[List[Tuple]]:
        if not path.startswith('/'):
            raise ImproperlyConfigured("Paths must start with '/'.")
        
        include_routes = kwargs.pop('include', None)
        if include_routes:
            RoutingHelpers._helper_inc_path(path, include_routes)
            pass
        else:
            response_model = kwargs.pop('response_model', None)
            strict_slashes = kwargs.pop('strict_slashes', True)
            methods = kwargs.pop('methods', None)

            if endpoint is None or not (inspect.iscoroutinefunction(endpoint) or inspect.isasyncgenfunction(endpoint)):
                raise ImproperlyConfigured("Invalid handler function provided for adding a route.")

            allowed_methods = {"GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS", "PATCH", "TRACE"}
            methods = methods or ["GET"]
            invalid_methods = [method for method in methods if method.upper() not in allowed_methods]
            if invalid_methods:
                raise ImproperlyConfigured(f"Invalid HTTP method(s) provided: {', '.join(invalid_methods)}")

            converted_path, path_regex = Converter()._regex_converter(path, strict_slashes, '')

            if DuplicateHandler._is_duplicate_route(converted_path, endpoint):
                raise ImproperlyConfigured("Duplicate endpoint detected for the same route.")

            _routes.append((
                converted_path,
                methods,
                endpoint,
                path_regex,
                response_model,
                endpoint
            ))
            _links.append((
                path,
                endpoint,
                kwargs.get('name')
            ))
            
            return (path, methods, endpoint, strict_slashes, response_model, endpoint, kwargs.get('name'))
        
    @staticmethod
    def re_rule(
        path_regex: str,
        endpoint: Optional[Callable[..., Awaitable[Any]]] = None,
        **kwargs: Any
    ) -> Optional[List[Tuple]]:
        if endpoint is None or not (inspect.iscoroutinefunction(endpoint) or inspect.isasyncgenfunction(endpoint)):
            raise ImproperlyConfigured("Invalid handler function provided for adding a route.")

        response_model = kwargs.pop('response_model', None)
        methods = kwargs.pop('methods', None)

        allowed_methods = {"GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS", "PATCH", "TRACE"}
        methods = methods or ["GET"]
        invalid_methods = [method for method in methods if method.upper() not in allowed_methods]
        if invalid_methods:
            raise ImproperlyConfigured(f"Invalid HTTP method(s) provided: {', '.join(invalid_methods)}")

        path_regex_compiled = re.compile(path_regex)

        if DuplicateHandler._is_duplicate_route_with_regex(path_regex_compiled.pattern, endpoint):
            raise ImproperlyConfigured("Duplicate endpoint detected for the same route.")

        _routes.append((
            path_regex,
            methods,
            endpoint,
            path_regex_compiled,
            response_model,
            endpoint
        ))
        _links.append(
            (
                path_regex,
                endpoint,
                kwargs.get('name', None)
            )
        )
        return (path_regex, methods, endpoint, kwargs.get('strict_slashes', False), response_model, endpoint, kwargs.get('name'))
    
    @staticmethod
    def rule_all(
        path: str,
        response_model: Optional[Type[T]] = None,
        endpoint: Callable[..., Awaitable[T]] = None,
        strict_slashes: bool = True,
        name: Optional[str] = None
    ) -> Callable[..., Awaitable[T]]:
        
        handler = endpoint
        allowed_methods = [
            "GET", "POST", "PUT", "DELETE",
            "HEAD", "OPTIONS", "PATCH", "TRACE"
        ]

        if not (inspect.iscoroutinefunction(handler) or inspect.isasyncgenfunction(handler)):
            raise TypeError("ASGI can only register asynchronous functions.")
        
        if not path.startswith('/'):
            raise ImproperlyConfigured("Paths must start with '/'.")
        
        DuplicateHandler._duplicate_handler(handler)

        converted_path, path_regex = Converter()._regex_converter(
            path, strict_slashes, ''
        )
        _routes.append(
            (
                converted_path,
                allowed_methods,
                handler,
                path_regex,
                response_model,
                endpoint
            )
        )
        
        _links.append((
            path,
            endpoint,
            name
        ))
        
        return (
            path,
            allowed_methods,
            handler,
            strict_slashes,
            response_model,
            endpoint,
            name
        )
        
class WebsocketRouting:
    @staticmethod
    def websocket(
        path: str,
        endpoint: Optional[Callable[..., Awaitable[T]]] = None,
        **kwargs: Any
    ) -> None:
        
        if not path.startswith('/'):
            raise ImproperlyConfigured("Websocket paths must start with '/'.")
        
        if 'include' in kwargs:
            websocket_include = kwargs.get('include', None)
            if websocket_include:
                RoutingHelpers._helper_inc_websockets(path, websocket_include)
                return None
        else:
            if endpoint is None:
                raise ImproperlyConfigured("Handler function is required for adding a websocket route.")

            if not inspect.iscoroutinefunction(endpoint) and not inspect.isasyncgenfunction(endpoint):
                raise ImproperlyConfigured("ASGI Websocket can only register asynchronous functions.")
            
            compiled_path, path_regex = Converter()._regex_converter(path, False)
            handler = endpoint
            _websockets.append((compiled_path, path_regex, handler))
            return (path, handler)

class _SchematicInstance:
    def __init__(self) -> None:
        self.schematic = None
        self.schematic_id = None

    def _process_routes(self, schematic_instance: [ASGIApp], url_prefix: str) -> None:
        for route in schematic_instance.routes:
            path, methods, handler, strict_slashes, response_model, endpoint = route

            if path.endswith('/'):
                path = ''  # Simulating the root_path

            converted_path, path_regex = Converter()._regex_converter(url_prefix + path, strict_slashes)

            _routes.append(
                (
                    converted_path,
                    methods,
                    handler,
                    path_regex,
                    response_model,
                    endpoint,
                )
            )
            _links.append((
                url_prefix + path,
                endpoint,
                schematic_instance.name
            ))

    def _process_schematic_instance(self, schematic_instance: [ASGIApp], url_prefix: str) -> None:
        if schematic_instance.schematic_id is not None:
            self._update_schematic_info(schematic_instance)
            self._print_schematic_info(schematic_instance, url_prefix)

        self._process_websockets(schematic_instance, url_prefix)

    def _update_schematic_info(self, schematic_instance: [ASGIApp]) -> None:
        self.schematic_id = schematic_instance.schematic_id
        self.schematic = schematic_instance.get_schematic()

    def _print_schematic_info(self, schematic_instance: [ASGIApp], url_prefix: str) -> None:
        schematic_name = schematic_instance.name
        url_prefix_info = (
            f"at {ColoursCode.BOLD}{ColoursCode.GREEN}{url_prefix}{ColoursCode.RESET} url-prefix"
            if url_prefix
            else "with no specific URL prefix"
        )

        serving_message = (
            f"\n Serving Schematic {ColoursCode.BOLD}{ColoursCode.GREEN}'{schematic_name}'{ColoursCode.RESET} Instance {url_prefix_info}."
        )
        instance_id_message = f"Instance-ID: {ColoursCode.BOLD}{ColoursCode.GREEN}{self.schematic_id}{ColoursCode.RESET}"

        print(serving_message)
        print(instance_id_message)

        paths = [route[0] for route in schematic_instance.routes]
        if paths:
            print(
                f"{ColoursCode.CYAN}Routes{ColoursCode.RESET}: {ColoursCode.GREEN}{paths}{ColoursCode.RESET} \n"
            )
        else:
            print(
                f"{ColoursCode.RED}No HTTP routes found for this schematic.{ColoursCode.RESET} \n"
            )

    def _process_websockets(self, schematic_instance: [ASGIApp], url_prefix: str) -> None:
        if schematic_instance.websockets:
            print(f"{ColoursCode.BOLD}WebSocket Routes:{ColoursCode.RESET}")

            for path, handler in schematic_instance.websockets:
                full_path = url_prefix + path
                _websockets.append(
                    (
                        full_path,
                        Converter()._regex_converter(full_path, False)[1],
                        handler
                    )
                )
                print(f"  - Path: {ColoursCode.GREEN}{full_path}{ColoursCode.RESET}")
                print(f"    Handler: {handler.__name__}")
            
            print()
        else:
            print(f"{ColoursCode.RED}No WebSocket routes found for this schematic.{ColoursCode.RESET}\n")

_schematic = _SchematicInstance()

def link(path: str, instance: Callable[..., Awaitable[Schematic]]) -> None:
    try:
        module_name, class_name = instance.rsplit('.', 1)
        module = import_module(module_name)
        schematic_instance = getattr(module, class_name)
        
        if not isinstance(schematic_instance, Schematic):
            raise TypeError(f"{instance} is not a valid Aquilify Schematic ASGIApp class")
            
        _schematic._process_routes(schematic_instance, path)
        _schematic._process_schematic_instance(schematic_instance, path)
    except (ValueError, AttributeError, ModuleNotFoundError) as e:
        raise ImportError(f"Failed to import {instance}: {e}")
    except TypeError as e:
        raise TypeError(f"Error processing {instance}: {e}")

class DynamicModuleLoader:
    class ModuleImportError(Exception):
        pass

    class RoutingVariableError(Exception):
        pass

    def __init__(self):
        pass

    def _module_loader(self, dotted_path: str) -> Optional[Any]:
        """
        Load a module given its dotted path.

        Args:
        - dotted_path (str): The dotted path of the module.

        Returns:
        - Optional[Any]: The loaded module if successful, otherwise None.

        Raises:
        - ModuleImportError: If unable to import the specified module.
        """
        try:
            return importlib.import_module(dotted_path)
        except ImportError as e:
            raise self.ModuleImportError(f"Unable to import module {dotted_path}: {e}")

    def _routing_avrs(self, module: Any, namespace: str = '') -> Optional[List[Any]]:
        """
        Retrieve the routing variable list from a module.

        Args:
        - module (Any): The module object.
        - namespace (str): The namespace for the routing variable.

        Returns:
        - Optional[List[Any]]: The routing variable list if found, otherwise None.

        Raises:
        - RoutingVariableError: If the routing variable is not found or not a list in the module.
        """
        routing_var = getattr(module, f"{namespace}ROUTER", None)
        if routing_var is None:
            raise self.RoutingVariableError(f"No '{namespace}ROUTER' variable found in the module")
        if not isinstance(routing_var, list):
            raise self.RoutingVariableError(f"'{namespace}ROUTER' variable is not a list in the module")
        return routing_var

    def _include(self, dotted_path: str, namespace: str = '') -> Optional[List[Any]]:
        """
        Include routing configuration from other sources.

        Args:
        - dotted_path (str): The dotted path of the module.
        - namespace (str): The application namespace.

        Returns:
        - Optional[List[Any]]: The included routing configuration if successful, otherwise None.

        Raises:
        - ModuleImportError: If unable to import the specified module.
        - RoutingVariableError: If the routing variable is not found or not a list in the module.
        """
        try:
            module: Any = self._module_loader(dotted_path)
            return self._routing_avrs(module, namespace)
        except (self.ModuleImportError, self.RoutingVariableError) as e:
            print(f"Error: {e}")
            return None

def include(args: str, namespace: str = '') -> Optional[List[Any]]:
    """
    Include routing configuration from specified modules.

    Args:
    - args (str): The modules' dotted paths.
    - namespace (str): The application namespace.

    Returns:
    - Optional[List[Any]]: The included routing configurations.

    Example:
    >>> include('module1.urls', namespace='my_namespace')
    """
    cls: Optional[List[Any]] = DynamicModuleLoader()._include(args, namespace)
    return cls


from aquilify.wrappers import Response, Request
from aquilify.utils.cache import cc_delim_re, get_conditional_response, set_response_etag
from aquilify.utils.http import parse_http_date_safe

class ConditionalGetMiddleware:
    def __init__(self) -> None:
        pass
    
    def needs_etag(self, response):
        cache_control_headers = cc_delim_re.split(response.headers.get("Cache-Control", ""))
        return all(header.lower() != "no-store" for header in cache_control_headers)
    
    async def __call__(self, request: Request, response: Response):

        if request.method != "GET":
            return response

        if self.needs_etag(response) and not response.headers.get("ETag"):
            set_response_etag(response)

        etag = response.headers.get("ETag")
        last_modified = response.headers.get("Last-Modified")
        last_modified = last_modified and parse_http_date_safe(last_modified)

        if etag or last_modified:
            return get_conditional_response(
                request,
                etag=etag,
                last_modified=last_modified,
                response=response,
            )

        return response



import importlib.util

class CORSConfigSettings:
    def __init__(self) -> None:
        self.settings_module_path: str = "./settings.py"

    def _fetch_sessions_config(self):
        try:
            settings = self._load_settings_module()

            allowed_origin = getattr(settings, "CORS_ALLOWED_ORIGIN", ["*"])
            allowed_method = getattr(settings, "CORS_ALLOWED_METHOD", ["*"])
            allowed_headers = getattr(settings, "CORS_ALLOWED_HEADERS", ["*"])
            allowed_credentials = getattr(settings, "CORS_ALLOWED_CREDENTIALS", False)
            expose_headers = getattr(settings, "CORS_EXPOSED_HEADERS", set())
            max_age = getattr(settings, "CORS_MAX_AGE", 600)
            security_headers = getattr(settings, "CORS_SECURITY_HEADERS", {})
            preflight_response = getattr(settings, "CORS_PREFLIGHT_RESPONSE", None)
            preflight_cache = getattr(settings, "CORS_PREFLIGHT_CACHE", True)
            origin_whitelist = getattr(settings, "CORS_ORIGIN_WHITELIST", None)
            automatic_preflight_handling = getattr(settings, "CORS_AUTOMATIC_PREFLIGHT_HANDLIGN", True)
            dynamic_headers_whitelist = getattr(settings, "CORS_DYNAMIC_HEADERS_WHITELIST", None)
            log_request = getattr(settings, "CORS_LOG_REQUEST", True)
            exclude_paths = getattr(settings, "CORS_EXCLUDE_PATHS", [])
            response_handler = self._fetch_response_handler_functions("CORS_RESPONSE_HANDLER")

            if security_headers is None:
                pass

            cors_settings = {
                "allowed_origin": allowed_origin,
                "allowed_method": allowed_method,
                "allowed_credentials": allowed_credentials,
                "allowed_headers": allowed_headers,
                "expose_headers": expose_headers,
                "max_age": max_age,
                "security_headers": security_headers,
                "preflight_cache": preflight_cache,
                "preflight_response": preflight_response,
                "origin_whitelist": origin_whitelist,
                "automatic_preflight_handling": automatic_preflight_handling,
                "dynamic_headers_whitelist": dynamic_headers_whitelist,
                "log_request": log_request,
                "response_handler": response_handler,
                "exclude_paths": exclude_paths
            }
            return cors_settings

        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")
        
    def _fetch_security_headers(self, settings):
        try:
            security_headers = getattr(settings, "CORS_SECURITY_HEADERS", {})
            return security_headers if security_headers else {}
        except AttributeError:
            return {}

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings
    
    def fetch(self):
        cors_setting = self._fetch_sessions_config()
        if cors_setting is None:
            raise KeyError("CORS Configuration missing in settings.py")
        return cors_setting
    
    def _fetch_response_handler_functions(self, object):
        settings = self._load_settings_module()
        
        if settings:
            path_to_callback = getattr(settings, object, None)
            
            if path_to_callback:
                try:
                    parts = path_to_callback.split('.')
                    module_path = '.'.join(parts[:-1])
                    callback_name = parts[-1]
                    callback_module = importlib.import_module(module_path)
                    callback_obj = getattr(callback_module, callback_name)
                    if isinstance(callback_obj, type):
                        callback_obj = callback_obj()
                    
                    return callback_obj
                except (AttributeError, ImportError, ValueError) as e:
                    print(f"Error fetching callback: {e}")
        
        return None



import typing

if typing.TYPE_CHECKING:
    from .wrappers import Request
    from .wrappers import Response
    from .wrappers import WebSocket
    
T = typing.TypeVar("T")

AppType = typing.TypeVar("AppType")

Scope = typing.MutableMapping[str, typing.Any]
Message = typing.MutableMapping[str, typing.Any]

Receive = typing.Callable[[], typing.Awaitable[Message]]
Send = typing.Callable[[Message], typing.Awaitable[None]]

ASGIApp = typing.Callable[[Scope, Receive, Send], typing.Awaitable[None]]

StatelessLifespan = typing.Callable[[AppType], typing.AsyncContextManager[None]]
StatefulLifespan = typing.Callable[
    [AppType], typing.AsyncContextManager[typing.Mapping[str, typing.Any]]
]
Lifespan = typing.Union[StatelessLifespan[AppType], StatefulLifespan[AppType]]

HTTPExceptionHandler = typing.Callable[
    ["Request", Exception], typing.Union["Response", typing.Awaitable["Response"]]
]
WebSocketExceptionHandler = typing.Callable[
    ["WebSocket", Exception], typing.Awaitable[None]
]
ExceptionHandler = typing.Union[HTTPExceptionHandler, WebSocketExceptionHandler]

RouteInfo = typing.Tuple[
    str, typing.List[str], typing.Callable[..., typing.Awaitable[T]], typing.List[str], typing.Pattern, typing.Type[T], str, typing.Dict[str, typing.Any]
]
Routes = typing.List[RouteInfo]
WebSocketRoute = typing.Tuple[str, typing.Callable[..., typing.Awaitable[T]]]
WebSocketRoutes = typing.List[WebSocketRoute]


def debug_404(links, data):
    html_code = f'''
    <!DOCTYPE html>
    <html lang="en">

    <head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>404 Error - Page Not Found</title>
    <style>
        *{{margin:0;padding:0;box-sizing:border-box;}}
        body{{font-family:Arial,sans-serif;background-color:#f4f4f4;display:flex;justify-content:center;align-items:center;height:100vh;}}
        .container{{width:80%;text-align:center;}}
        h1{{font-size:4em;margin-bottom:20px;color:#333;}}
        .error-details{{background-color:#FFFACD;padding:30px;border-radius:8px;box-shadow:0 0 10px rgba(0,0,0,0.1);}}
        .debugging-info{{text-align:left;margin-top:30px;}}
        .debugging-info h2{{font-size:2em;margin-bottom:15px;color:#333;}}
        .debugging-info p{{margin-bottom:10px;font-size:1.2em;}}
        .debugging-info p span{{font-weight:bold;}}
        .routing-pattern-box,.warning-box{{margin-top:20px;padding:15px;border-radius:5px;}}
        .routing-pattern-box{{background-color:#FFFACD;border:1px solid black;}}
        .routing-pattern-box h3,.warning-box h3{{font-size:1.5em;margin-bottom:10px;color:black;}}
        .routing-pattern-box ul{{list-style:none;padding:0;text-align:left;}}
        .routing-pattern-box ul li{{margin-bottom:8px;color:#555;}}
        .warning-box{{background-color:#ffeaea;border:1px solid #e74c3c;}}
        .warning-box p{{color:#e74c3c;font-weight:bold;font-size:1.2em;}}
        .warning-box p:before{{content:"⚠ ";}}
        p.message{{margin-top:30px;font-size:1.1em;color:#555;}}
        @media only screen and (max-width:768px){{h1{{font-size:3em;}}.debugging-info h2{{font-size:1.5em;}}.debugging-info p{{font-size:1em;}}.routing-pattern-box h3,.warning-box h3{{font-size:1.2em;}}.warning-box p{{font-size:1em;}}p.message{{font-size:0.9em;}}}}
    </style>
    </head>

    <body>
    <div class="container">
        <h1>404 - Page Not Found</h1>
        <div class="error-details">
        <div class="debugging-info">
            <h2>Debugging Details:</h2>
            <p>Requested Method: <span>{data['method']}</span></p>
            <p>Requested URL: <span>{data['url']}</span></p>
            <p>User's IP Address: <span>{data['client_ip']}</span></p>
            <p>User Agent: <span>{data['user_agent']}</span></p>
            <div class="routing-pattern-box">
            <h3>Routing Pattern:</h3>
            <ul>
                {''.join([f'<b><li>Name: {name} | | Path: {path}</li></b>' for path, _, name in links])}
            </ul>
            </div>
            <div class="warning-box">
            <h3>Warning:</h3>
            <p>Current path "{data['path']}" not found in Routing Pattern!</p>
            </div>
            <p class="message">You're seeing this error because you have DEBUG = True in Aquilify settings. Change that to false to display the standard 404 error page.</p>
        </div>
        </div>
    </div>
    </body>

    </html>
    '''
    return html_code

def debug_405(data):
    html_code = f'''
    <!DOCTYPE html>
    <html lang="en">

    <head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>405 Error - Method Not Allowed</title>
    <style>
        *{{margin:0;padding:0;box-sizing:border-box;}}
        body{{font-family:Arial,sans-serif;background-color:#f4f4f4;display:flex;justify-content:center;align-items:center;height:100vh;}}
        .container{{width:80%;text-align:center;}}
        h1{{font-size:4em;margin-bottom:20px;color:#333;}}
        .error-details{{background-color:#FFFACD;padding:30px;border-radius:8px;box-shadow:0 0 10px rgba(0,0,0,0.1);}}
        .debugging-info{{text-align:left;margin-top:30px;}}
        .debugging-info h2{{font-size:2em;margin-bottom:15px;color:#333;}}
        .debugging-info p{{margin-bottom:10px;font-size:1.2em;}}
        .debugging-info p span{{font-weight:bold;}}
        .warning-box{{background-color:#ffeaea;border:1px solid #e74c3c;padding:15px;border-radius:5px;}}
        .warning-box h3{{font-size:1.5em;margin-bottom:10px;color:#e74c3c;}}
        .warning-box p{{color:#e74c3c;font-weight:bold;font-size:1.2em;}}
        .warning-box p:before{{content:"⚠ ";}}
        p.message{{margin-top:30px;font-size:1.1em;color:#555;}}
        @media only screen and (max-width:768px){{h1{{font-size:3em;}}.debugging-info h2{{font-size:1.5em;}}.debugging-info p{{font-size:1em;}}.warning-box h3{{font-size:1.2em;}}.warning-box p{{font-size:1em;}}p.message{{font-size:0.9em;}}}}
    </style>
    </head>

    <body>
    <div class="container">
        <h1>405 - Method Not Allowed</h1>
        <div class="error-details">
        <div class="debugging-info">
            <h2>Debugging Details:</h2>
            <p>Requested Method: <span>{data['method']}</span></p>
            <p>Requested URL: <span>{data['url']}</span></p>
            <p>Allowed Methods : <span>{data['allowed_method']}</span></p>
            <p>User's IP Address: <span>{data['client_ip']}</span></p>
            <p>User Agent: <span>{data['user_agent']}</span></p>
            <div class="warning-box">
            <h3>Warning:</h3>
            <p>Method not allowed for the requested URL!</p>
            </div>
            <p class="message">You're seeing this error because you have DEBUG = True in Aquilify settings. Change that to false to display the standard 405 error page.</p>
        </div>
        </div>
    </div>
    </body>

    </html>
    '''

    return html_code



import datetime
import logging
import os
from logging.handlers import TimedRotatingFileHandler
from typing import Optional

from ..settings.logger import LoggingConfigSettings

_settings = LoggingConfigSettings().fetch()

class LoggingMiddleware:
    def __init__(
        self
    ):
        self.log_file_path = _settings.get('file_path') or None
        self.log_user_agent = _settings.get('user_agent') or True
        self.log_client_ip = _settings.get('client_ip') or True
        self.max_log_size = _settings.get('log_size') or 10 * 1024 * 1024
        self.backup_count = _settings.get('backup_count') or 5
        self.log_level = _settings.get('log_level') or logging.INFO
        self.output_stream = _settings.get('output_stream') or 'file'
        self.append_logs = _settings.get('append_logs') or False
        self.timestamp_format = _settings.get('timestamp_format') or "%Y-%m-%d %H:%M:%S"
        self.log_response_time = _settings.get('log_response_time') or True
        self.url_patterns_to_log = _settings.get('url_patterns_to_log') or []

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)

        self._setup_logging(_settings.get('log_format') or None)

    def _setup_logging(self, log_format: Optional[str]):
        log_formatter = log_format or '%(asctime)s - %(levelname)s - %(message)s'

        if self.log_file_path and self.output_stream == 'file':
            if not os.path.exists(os.path.dirname(self.log_file_path)):
                os.makedirs(os.path.dirname(self.log_file_path))

            if self.append_logs:
                file_handler = logging.FileHandler(self.log_file_path)
            else:
                file_handler = TimedRotatingFileHandler(
                    self.log_file_path,
                    when='midnight',  # Rotate logs at midnight
                    interval=1,  # Rotate daily
                    backupCount=self.backup_count
                )
            file_handler.setFormatter(logging.Formatter(log_formatter))
            self.logger.addHandler(file_handler)
        else:
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(logging.Formatter(log_formatter))
            self.logger.addHandler(console_handler)

    async def log_request(self, request):
        if not self._should_log_request(request):
            return

        start_time = datetime.datetime.now()
        log_entry = self._create_log_entry(request)

        if self.log_response_time:
            request.context['start_time'] = start_time
            request.context['log_entry'] = log_entry

    async def log_response(self, response, context):
        log_entry = context.get('log_entry')
        start_time = context.get('start_time')

        if log_entry and start_time:
            log_entry['status_code'] = response.status_code
            end_time = datetime.datetime.now()
            response_time = (end_time - start_time).total_seconds() * 1000  # milliseconds
            log_entry['response_time_ms'] = response_time

            log_message = self._format_log_entry(log_entry)
            self.logger.info(log_message)

    def _should_log_request(self, request):
        if not self.url_patterns_to_log:
            return True

        for pattern in self.url_patterns_to_log:
            if pattern in request.url:
                return True
        return False

    def _create_log_entry(self, request):
        log_entry = {
            'timestamp': datetime.datetime.now().strftime(self.timestamp_format),
            'method': request.method,
            'url': request.url,
        }

        if self.log_user_agent:
            log_entry['user_agent'] = request.headers.get('user-agent')

        if self.log_client_ip:
            log_entry['client_ip'] = request.client

        return log_entry

    def _format_log_entry(self, log_entry):
        formatted_log = f"[{log_entry['timestamp']}] - {log_entry['method']} {log_entry['url']}"
        if 'status_code' in log_entry:
            formatted_log += f" - Status Code: {log_entry['status_code']}"
        if 'user_agent' in log_entry:
            formatted_log += f" - User Agent: {log_entry['user_agent']}"
        if 'client_ip' in log_entry:
            formatted_log += f" - Client IP: {log_entry['client_ip']}"
        if 'response_time_ms' in log_entry:
            formatted_log += f" - Response Time: {log_entry['response_time_ms']} ms"
        return formatted_log

    async def __call__(self, request, response):
        context = {
            'request': request,
        }
        await self.log_request(request)
        await self.log_response(response, request.context)
        return response



from contextlib import AbstractContextManager
import sqlite3
from enum import Enum

from .exceptions import SessionExecuteError

class IsolationLevel(Enum):
    DEFERRED = "DEFERRED"
    IMMEDIATE = "IMMEDIATE"
    EXCLUSIVE = "EXCLUSIVE"

class TransactionContextManager(AbstractContextManager):
    def __init__(self, db_connection, isolation_level=IsolationLevel.DEFERRED):
        self._db_connection = db_connection
        self._isolation_level = isolation_level
        self._nested = 0
        self._completed = False

    def __enter__(self):
        self._nested += 1
        if self._nested == 1:
            self._db_connection.execute(f"BEGIN {self._isolation_level.value} TRANSACTION")
        else:
            self._db_connection.execute(f"SAVEPOINT sp_{self._nested}")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._nested -= 1
        if self._nested == 0:
            if exc_type is None and not self._completed:
                self.commit()
            else:
                self.rollback()

    def commit(self):
        """Manually commit the transaction."""
        if not self._completed:
            try:
                self._db_connection.execute("COMMIT")
                self._completed = True
            except sqlite3.Error as e:
                raise SessionExecuteError(f"Error committing transaction: {e}") from e

    def rollback(self):
        """Manually roll back the transaction."""
        if not self._completed:
            try:
                self._db_connection.execute("ROLLBACK")
                self._completed = True
            except sqlite3.Error as e:
                raise SessionExecuteError(f"Error rolling back transaction: {e}") from e

    def __call__(self, isolation_level=IsolationLevel.DEFERRED):
        """Allows chaining transaction context managers."""
        return TransactionContextManager(self._db_connection, isolation_level)



from __future__ import annotations

import json
import secrets
import gzip 
import io

from typing import Optional, List, Union, Callable, Type, Any, Dict
from datetime import datetime, timedelta
from ..exception.__handler import handle_exception
from ._http_status import HTTP_STATUS_PHRASE

class HTTPStatus:
    def __init__(self, code: int, phrase: str):
        self.code = code
        self.phrase = phrase

    def __str__(self):
        return f"{self.code} {self.phrase}"

class Response:
    def __init__(
        self,
        content: Union[str, bytes, Callable, None] = None,
        status_code: int = 200,
        headers: Optional[Dict[bytes, bytes]] = None,
        content_type: str = "text/plain",
        compress: bool = False,
    ):
        self.status_code = status_code
        self.headers = headers or {}
        self.content_type = content_type
        self.content = content
        self.streaming = callable(content)
        self.compress = compress
        self.encoding = 'utf-8'

    async def __call__(self, scope, receive, send):
        response_headers = {
            b'Content-Type': f"{self.content_type}; charset={self.encoding}".encode(),
            **self.headers,
        }

        response_headers = {key.encode() if isinstance(key, str) else key: value.encode() if isinstance(value, str) else value for key, value in response_headers.items()}

        try:
            content_length = 0
            if self.content:
                if isinstance(self.content, str):
                    content_length = len(self.content.encode('utf-8'))
                elif isinstance(self.content, bytes):
                    content_length = len(self.content)
                elif callable(self.content):
                    content_length = await self.get_stream_content_length(scope, receive, send)

            response_headers[b'Content-Length'] = str(content_length).encode()
            
            if self.compress:
                response_headers[b'Content-Encoding'] = b'gzip'
                await send({
                    'type': 'http.response.start',
                    'status': self.status_code,
                    'headers': list(response_headers.items()),
                })

                if self.streaming:
                    await self._send_streaming_response_compressed(scope, receive, send)
                else:
                    await self._send_standard_response_compressed(send)
            else:
                # No compression
                await send({
                    'type': 'http.response.start',
                    'status': self.status_code,
                    'headers': list(response_headers.items()),
                })

                if self.streaming:
                    await self._send_streaming_response(scope, receive, send)
                else:
                    await self._send_standard_response(send)

        except Exception as e:
            await handle_exception(e)

    async def _send_streaming_response_compressed(self, scope, receive, send):
        try:
            if callable(self.content):
                gzip_buffer = io.BytesIO()
                gzip_stream = gzip.GzipFile(fileobj=gzip_buffer, mode='w')
                async for chunk in self.content(scope, receive, send):
                    gzip_stream.write(chunk)
                    await send({
                        'type': 'http.response.body',
                        'body': gzip_buffer.getvalue(),
                        'more_body': True,
                    })
                    gzip_buffer.truncate(0)
                    gzip_buffer.seek(0)

                gzip_stream.close()
        except Exception as e:
            await handle_exception(e)

    async def _send_standard_response_compressed(self, send):
        try:
            if self.content is not None:
                if isinstance(self.content, str):
                    compressed_content = gzip.compress(self.content.encode('utf-8'))
                    await send({
                        'type': 'http.response.body',
                        'body': compressed_content,
                    })
                elif isinstance(self.content, bytes):
                    compressed_content = gzip.compress(self.content)
                    await send({
                        'type': 'http.response.body',
                        'body': compressed_content,
                    })
                else:
                    # Default to JSON serialization for other types
                    content_str = json.dumps(self.content)
                    compressed_content = gzip.compress(content_str.encode('utf-8'))
                    await send({
                        'type': 'http.response.body',
                        'body': compressed_content,
                    })

        except Exception as e:
            await handle_exception(e)

    async def _send_streaming_response(self, scope, receive, send):
        try:
            if self.content:
                async for chunk in self.content(scope, receive, send):
                    await send({
                        'type': 'http.response.body',
                        'body': chunk.encode('utf-8') if isinstance(chunk, str) else chunk,
                    })

        except Exception as e:
            await handle_exception(e)

    async def _send_standard_response(self, send):
        try:
            if self.content is not None:
                if isinstance(self.content, str):
                    await send({
                        'type': 'http.response.body',
                        'body': self.content.encode('utf-8'),
                    })
                elif isinstance(self.content, bytes):
                    await send({
                        'type': 'http.response.body',
                        'body': self.content,
                    })
                else:
                    # Default to JSON serialization for other types
                    content_str = json.dumps(self.content)
                    await send({
                        'type': 'http.response.body',
                        'body': content_str.encode('utf-8'),
                    })

        except Exception as e:
            await handle_exception(e)

    async def get_stream_content_length(self, scope, receive, send):
        content_length = 0
        async for chunk in self.content(scope, receive, send):
            if isinstance(chunk, str):
                content_length += len(chunk.encode('utf-8'))
            elif isinstance(chunk, bytes):
                content_length += len(chunk)
        return content_length

    @property
    def max_cookie_size(self):
        return int(self.headers.get("Set-Cookie-Size", 4096))

    async def set_cookie(
        self,
        key: str,
        value: str,
        max_age: int = None,
        expires: Optional[Union[int, datetime]] = None,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Optional[str] = None,
    ):
        try:
            cookie_parts = [f"{key}={value}"]

            if max_age is not None:
                cookie_parts.append(f"Max-Age={max_age}")
            if expires is not None:
                if isinstance(expires, int):
                    expires = datetime.now() + timedelta(seconds=expires)
                cookie_parts.append(f"Expires={expires.strftime('%a, %d %b %Y %H:%M:%S GMT')}")
            if path:
                cookie_parts.append(f"Path={path}")
            if domain:
                cookie_parts.append(f"Domain={domain}")
            if secure:
                cookie_parts.append("Secure")
            if httponly:
                cookie_parts.append("HttpOnly")
            if samesite:
                cookie_parts.append(f"SameSite={samesite}")

            self.headers["Set-Cookie"] = "; ".join(cookie_parts)

        except Exception as e:
            await handle_exception(e)
            
    def _set_cookie(
        self,
        key: str,
        value: str = None,
        max_age: int = None,
        expires: Optional[Union[int, datetime]] = None,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Optional[str] = None,
    ):
        try:
            cookie_parts = [f"{key}={value}"]

            if max_age is not None:
                cookie_parts.append(f"Max-Age={max_age}")
            if expires is not None:
                if isinstance(expires, int):
                    expires = datetime.now() + timedelta(seconds=expires)
                cookie_parts.append(f"Expires={expires.strftime('%a, %d %b %Y %H:%M:%S GMT')}")
            if path:
                cookie_parts.append(f"Path={path}")
            if domain:
                cookie_parts.append(f"Domain={domain}")
            if secure:
                cookie_parts.append("Secure")
            if httponly:
                cookie_parts.append("HttpOnly")
            if samesite:
                cookie_parts.append(f"SameSite={samesite}")

            self.headers["Set-Cookie"] = "; ".join(cookie_parts)

        except Exception as e:
            raise e

    async def delete_cookie(self, key: str):
        try:
            # Create a new dictionary for the Set-Cookie header with Expires in the past
            expires = datetime(1970, 1, 1).strftime('%a, %d %b %Y %H:%M:%S GMT')
            self.headers["Set-Cookie"] = f"{key}=; Expires={expires}; Max-Age=0; Path=/"
        except Exception as e:
            await handle_exception(e)
            
    def _del_cookie(self, key, path="/", domain=None, samesite=None):
        # Browsers can ignore the Set-Cookie header if the cookie doesn't use
        # the secure flag and:
        # - the cookie name starts with "__Host-" or "__Secure-", or
        # - the samesite is "none".
        secure = key.startswith(("__Secure-", "__Host-")) or (
            samesite and samesite.lower() == "none"
        )
        self._set_cookie(
            key,
            max_age=0,
            path=path,
            domain=domain,
            secure=secure,
            expires=datetime(1970, 1, 1),
            samesite=samesite,
        )

    async def get_cookie(self, key: str) -> Optional[str]:
        try:
            cookies = self.headers.get("Cookie")
            if cookies:
                for cookie in cookies.split("; "):
                    name, value = cookie.split("=")
                    if name == key:
                        return value
            return None

        except Exception as e:
            await handle_exception(e)

    async def generate_secure_cookie(self, key: str, max_age: int = 3600) -> str:
        try:
            secure_token = secrets.token_hex(32)

            # Store the token in a cookie
            await self.set_cookie(key, secure_token, max_age=max_age, secure=True, httponly=True, samesite="Strict")

            return secure_token

        except Exception as e:
            await handle_exception(e)

    async def verify_secure_cookie(self, key: str, token: str) -> bool:
        try:
            stored_token = await self.get_cookie(key)
            return stored_token == token

        except Exception as e:
            await handle_exception(e)

    async def last_modified(self, date: datetime):
        try:
            self.headers["Last-Modified"] = date.strftime("%a, %d %b %Y %H:%M:%S GMT")
        except Exception as e:
            await handle_exception(e)

    async def etag(self, value: str):
        try:
            self.headers["ETag"] = value
        except Exception as e:
            await handle_exception(e)

    async def cache_for(
        self,
        seconds: int,
        public: bool = True,
        immutable: bool = False,
    ):
        try:
            directives = []
            if public:
                directives.append("public")
            if immutable:
                directives.append("immutable")
            directives.append(f"max-age={seconds}")
            self.cache_control({"no-store": False, "private": False, "no-cache": False, "max-age": seconds})
            self.expires(datetime.now() + timedelta(seconds=seconds))

        except Exception as e:
            await handle_exception(e)

    async def no_cache(self):
        try:
            self.cache_control({"no-store": True, "private": False, "no-cache": True})
        except Exception as e:
            await handle_exception(e)

    async def content_disposition(self, filename: str, inline: bool = False):
        try:
            disposition = "inline" if inline else "attachment"
            self.headers["Content-Disposition"] = f'{disposition}; filename="{filename}"'
        except Exception as e:
            await handle_exception(e)

    async def json(self, content: Any, status_code: int = 200):
        try:
            self.content_type = "application/json"
            self.status_code = status_code
            self.content = content
            self.streaming = False
        except Exception as e:
            await handle_exception(e)

    async def stream(self, content: Union[str, bytes, Callable] = None):
        try:
            self.streaming = True
            if content is not None:
                self.content = content
        except Exception as e:
            await handle_exception(e)

    @property
    def status_text(self):
        return f"{self.status_code} {HTTP_STATUS_PHRASE(self.status_code, 'Unknown')}"

    async def vary(self, headers: List[str]):
        try:
            self.headers["Vary"] = ", ".join(headers)
        except Exception as e:
            await handle_exception(e)

    async def calculate_content_length(self):
        try:
            if self.content is not None:
                if isinstance(self.content, str):
                    return len(self.content.encode("utf-8"))
                elif isinstance(self.content, bytes):
                    return len(self.content)
                else:
                    # Default to JSON serialization for other types
                    content_str = json.dumps(self.content)
                    return len(content_str.encode("utf-8"))
            return 0
        except Exception as e:
            await handle_exception(e)

    async def date(self, date_str: str):
        try:
            self.headers["Date"] = date_str
        except Exception as e:
            await handle_exception(e)

    async def expires(self, expires_str: str):
        try:
            self.headers["Expires"] = expires_str
        except Exception as e:
            await handle_exception(e)
            
    @property
    def location(self):
        return self.headers.get("Location")

    @location.setter
    async def location(self, value):
        try:
            self.headers["Location"] = value
        except Exception as e:
            await handle_exception(e)

    async def accept_ranges(self, ranges: str):
        try:
            self.headers["Accept-Ranges"] = ranges
        except Exception as e:
            await handle_exception(e)

    async def age(self, age: int):
        try:
            self.headers["Age"] = str(age)
        except Exception as e:
            await handle_exception(e)

    async def allow(self, methods: List[str]):
        try:
            self.headers["Allow"] = ", ".join(methods)
        except Exception as e:
            await handle_exception(e)

    async def cache_control(self, directives: Dict[str, Union[str, int]]):
        try:
            directives_str = ", ".join([f"{key}={value}" for key, value in directives.items()])
            self.headers["Cache-Control"] = directives_str
        except Exception as e:
            await handle_exception(e)

    async def access_control_allow_credentials(self, allow: bool):
        try:
            self.headers["Access-Control-Allow-Credentials"] = str(allow).lower()
        except Exception as e:
            await handle_exception(e)

    async def access_control_allow_headers(self, headers: List[str]):
        try:
            self.headers["Access-Control-Allow-Headers"] = ", ".join(headers)
        except Exception as e:
            await handle_exception(e)

    async def access_control_allow_methods(self, methods: List[str]):
        try:
            self.headers["Access-Control-Allow-Methods"] = ", ".join(methods)
        except Exception as e:
            await handle_exception(e)

    async def access_control_allow_origin(self, origin: str):
        try:
            self.headers["Access-Control-Allow-Origin"] = origin
        except Exception as e:
            await handle_exception(e)

    async def access_control_expose_headers(self, headers: List[str]):
        try:
            self.headers["Access-Control-Expose-Headers"] = ", ".join(headers)
        except Exception as e:
            await handle_exception(e)

    async def add_etag(self):
        try:
            if 'ETag' not in self.headers:
                self.etag(self.get_etag())
        except Exception as e:
            await handle_exception(e)

    async def autocorrect_location_header(self):
        try:
            if self.status_code in {301, 302, 303, 307, 308} and 'Location' not in self.headers:
                self.location("/")
        except Exception as e:
            await handle_exception(e)

    async def automatically_set_content_length(self):
        try:
            if 'Content-Length' not in self.headers:
                content_length = await self.calculate_content_length()
                if content_length > 0:
                    self.headers['Content-Length'] = str(content_length)
        except Exception as e:
            await handle_exception(e)

    async def content_security_policy(self, policy: str):
        try:
            self.headers["Content-Security-Policy"] = policy
        except Exception as e:
            await handle_exception(e)

    async def content_security_policy_report_only(self, policy: str):
        try:
            self.headers["Content-Security-Policy-Report-Only"] = policy
        except Exception as e:
            await handle_exception(e)

    async def cross_origin_embedder_policy(self, policy: str):
        try:
            self.headers["Cross-Origin-Embedder-Policy"] = policy
        except Exception as e:
            await handle_exception(e)

    async def cross_origin_opener_policy(self, policy: str):
        try:
            self.headers["Cross-Origin-Opener-Policy"] = policy
        except Exception as e:
            await handle_exception(e)

    async def get_etag(self) -> str:
        try:
            return self.headers.get("ETag", "")
        except Exception as e:
            await handle_exception(e)

    async def implicit_sequence_conversion(self, enabled: bool):
        try:
            self.headers["Implicit-Sequence-Conversion"] = str(enabled).lower()
        except Exception as e:
            await handle_exception(e)

    @property
    def max_cookie_size(self) -> int:
        return int(self.headers.get("Set-Cookie-Size", 4096))

    async def www_authenticate(self, value: str):
        try:
            self.headers["WWW-Authenticate"] = value
        except Exception as e:
            await handle_exception(e)

    async def make_conditional(self, is_conditional: bool = True):
        try:
            if is_conditional:
                self.headers["Cache-Control"] = "no-cache"
        except Exception as e:
            await handle_exception(e)




from .base import BaseStorage
from .cookie import CookieStorage
from .session import SessionStorage


class FallbackStorage(BaseStorage):
    storage_classes = (CookieStorage, SessionStorage)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.storages = [
            storage_class(*args, **kwargs) for storage_class in self.storage_classes
        ]
        self._used_storages = set()

    def _get(self, *args, **kwargs):
        all_messages = []
        for storage in self.storages:
            messages, all_retrieved = storage._get()
            if messages is None:
                break
            if messages:
                self._used_storages.add(storage)
            all_messages.extend(messages)
            if all_retrieved:
                break
        return all_messages, all_retrieved

    def _store(self, messages, response, *args, **kwargs):
        for storage in self.storages:
            if messages:
                messages = storage._store(messages, response, remove_oldest=False)
            elif storage in self._used_storages:
                storage._store([], response)
                self._used_storages.remove(storage)
        return messages



from .request import Request as Request
from .response import Response as Response
from .websocket import (
    WebSocket as WebSocket,
    WebSocketDisconnect as WebSocketDisconnect,
    WebSocketState as WebSocketState
)
from .channels import Channel as Channel
from .reqparser import Reqparser as Parser
from .reqparser import ReqparserError as ReqparserError


import os
import gzip
import json
import logging
import uuid

from io import BytesIO
from ..settings.compression import CompressionSetting

class GzipMiddleware:
    def __init__(
        self
    ):
        self.compress_level = 6
        self.content_types = [
            'text/html',
            'text/css',
            'application/javascript',
            'application/json',
        ]
        self.ignore_content_length = None
        self.exclude_paths = []
        self.encodings = ['gzip']
        self.custom_compress_funcs = {}
        self.compression_folder = '.aquilify/compression/'
        self.config_file = '.aquilify/config.json'
        self.logger = self._setup_logger()
        self._create_aquilify_folder()
        self._create_compression_folder()

        self.settings = CompressionSetting._compression_settings()
        self.compress_level = self.settings['GZIP_COMPRESSION_LEVEL']
        self.content_types = self.settings['GZIP_COMPRESSION_CONTENT_TYPES']
        self.ignore_content_length = self.settings['GZIP_IGNORE_CONTENT_LENGHT']
        self.exclude_paths = self.settings['GZIP_EXCLUDE_PATHS']
        self.encodings = self.settings['GZIP_CONTENT_ENCODING']
        self.custom_compress_funcs = self.settings['GZIP_COMPRESSION_FUNCTION']

    def _setup_logger(self):
        logger = logging.getLogger('GzipMiddleware')
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(formatter)

        logger.addHandler(ch)
        return logger

    def _create_aquilify_folder(self):
        try:
            aquilify_folder = os.path.join(os.getcwd(), '.aquilify')
            if not os.path.exists(aquilify_folder):
                os.makedirs(aquilify_folder)
        except Exception as e:
            self.logger.error(f"Error creating '.aquilify' folder: {e}")

    def _create_compression_folder(self):
        try:
            compression_folder_path = os.path.join(os.getcwd(), self.compression_folder)
            if not os.path.exists(compression_folder_path):
                os.makedirs(compression_folder_path)
        except Exception as e:
            self.logger.error(f"Error creating '{self.compression_folder}' folder: {e}")

    async def __call__(self, request, response):
        try:
            if len(response.content) <= 50:
                return response
            
            if self._should_exclude_path(request.scope['path']):
                return response

            should_compress, content_type = self._should_compress(response)
            if should_compress:
                return response

            compressed_content = await self._compress_content(response.content, content_type)
            response.content = compressed_content
            response.headers.update(self._create_gzip_headers(compressed_content, content_type))

            self._save_compressed_data(compressed_content)
            self._update_config_file()
            self._cleanup_old_files()

        except Exception as e:
            self.logger.error(f"Error compressing response content: {e}")

        return response
    
    def _should_exclude_path(self, path):
        return any(path.startswith(excluded_path) for excluded_path in self.exclude_paths)

    def _should_compress(self, response):
        content_type = response.headers.get('Content-Type', '')
        encoding = response.headers.get('Content-Encoding', '')
        
        should_compress = (
            'Content-Encoding' not in response.headers and
            response.content and
            any(content_type.startswith(ct) for ct in self.content_types) and
            any(enc in self.encodings for enc in encoding.split(','))
        )
        return should_compress, content_type

    async def _compress_content(self, content, content_type):
        if content_type in self.custom_compress_funcs:
            return await self.custom_compress_funcs[content_type](content)

        if isinstance(content, (str, dict, list, tuple)):
            content = str(content).encode()
        elif not isinstance(content, bytes):
            raise ValueError("Unsupported response content type for compression")

        buffer = BytesIO()
        with gzip.GzipFile(fileobj=buffer, mode='w', compresslevel=self.compress_level) as gz:
            gz.write(content)

        return buffer.getvalue()

    def _create_gzip_headers(self, compressed_content, content_type):
        headers = {
            'Content-Encoding': 'gzip',
            'Vary': 'Accept-Encoding'
        }
        if not self.ignore_content_length:
            headers['Content-Length'] = str(len(compressed_content))

        if content_type:
            headers['Content-Type'] = content_type

        return headers

    def _save_compressed_data(self, compressed_content):
        try:
            filename = self._generate_unique_filename()
            file_path = os.path.join(os.getcwd(), self.compression_folder, filename)
            with open(file_path, 'wb') as file:
                file.write(compressed_content)
        except Exception as e:
            self.logger.error(f"Error saving compressed data: {e}")

    def _generate_unique_filename(self):
        return str(uuid.uuid4()) + '.gzip'

    def _update_config_file(self):
        try:
            config_path = os.path.join(os.getcwd(), self.config_file)
            if not os.path.exists(config_path):
                config_data = {"$compression": []}
            else:
                with open(config_path, 'r') as file:
                    config_data = json.load(file)

            filename = self._generate_unique_filename()
            config_data["$compression"].append(filename)

            with open(config_path, 'w') as file:
                json.dump(config_data, file, indent=4)
        except Exception as e:
            self.logger.error(f"Error updating config file: {e}")

    def _cleanup_old_files(self):
        try:
            compression_folder_path = os.path.join(os.getcwd(), self.compression_folder)
            files_in_compression_folder = os.listdir(compression_folder_path)

            if len(files_in_compression_folder) > 10: 
                oldest_files = sorted(files_in_compression_folder, key=lambda x: os.path.getmtime(os.path.join(compression_folder_path, x)))
                for file in oldest_files[:-10]: 
                    file_path = os.path.join(compression_folder_path, file)
                    if os.path.exists(file_path):
                        os.remove(file_path)
                    else:
                        self.logger.warning(f"File {file_path} does not exist.")
        except Exception as e:
            self.logger.error(f"Error cleaning up old files: {e}")



"""
Backend for test environment.
"""

from ... import mail
from .base import BaseEmailBackend


class EmailBackend(BaseEmailBackend):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not hasattr(mail, "outbox"):
            mail.outbox = []

    def send_messages(self, messages):
        """Redirect messages to the dummy outbox"""
        msg_count = 0
        for message in messages:
            message.message()
            mail.outbox.append(message)
            msg_count += 1
        return msg_count



import secrets
import logging
import asyncio

from datetime import datetime, timedelta

from aquilify.core import signing

from ..wrappers import Request, Response

from ..settings.csrf import CSRFConfigSettings

try:
    _config_settings = CSRFConfigSettings()
    _config_settings.fetch()
    _settings = _config_settings.csrf_data[0]
except Exception as e:
    raise Exception("CSRF Configuration Not Found in settings.py, configure the CSRF settings before you use.")

class CSRFTokenError(Exception):
    pass

class RateLimiter:
    def __init__(self, max_requests: int, time_window: timedelta):
        self.max_requests = max_requests
        self.time_window = time_window
        self.tokens = {}
    
    def check(self, client_ip: str) -> bool:
        current_time = datetime.utcnow()
        self.cleanup_expired_tokens(current_time)

        if client_ip not in self.tokens:
            self.tokens[client_ip] = {"count": 1, "start_time": current_time}
            return True

        if self.tokens[client_ip]["count"] < self.max_requests:
            self.tokens[client_ip]["count"] += 1
            return True

        return False

    def cleanup_expired_tokens(self, current_time):
        expired_ips = [ip for ip, data in self.tokens.items() if (current_time - data["start_time"]).total_seconds() > self.time_window.total_seconds()]
        for ip in expired_ips:
            del self.tokens[ip]

class CSRF:
    def __init__(
        self
    ):
        self.secret_key = _settings['options'].get('secret_key')
        self.expiration = _settings['options'].get('expiration') or timedelta(hours=1)
        self.csrf_token_key = _settings['options'].get('token_key') or "_csrf_token"
        self.cookie_options = _settings['options'].get('cookie_options') or {"httponly": True, "secure": True, "samesite": 'strict'}
        self.protected_methods = _settings['options'].get('protected_methods') or ["POST", "PUT", "DELETE"]
        self.logger = _settings['options'].get('logger') or logging.getLogger(__name__)
        self.ip_verification = _settings['options'].get('ip_verification') or True
        self.trusted_ips = set(_settings['options'].get('trusted_ips')) if _settings['options'].get('trusted_ips') else set()
        self.rate_limit = None
        self.rate_limiter = RateLimiter(max_requests=self.rate_limit["max_requests"], time_window=self.rate_limit["time_window"]) if self.rate_limit else None
        self.max_token_lifetime = _settings['options'].get('max_token_lifetime') or timedelta(days=7)
        self.token_refresh_interval = _settings['options'].get('token_refresh_interval') or timedelta(days=6)
        self.serializer = signing
        self.revoked_tokens = set()

        assert _settings.get('backend') == 'aquilify.security.csrf.CSRF', "Invalid backend for CSRF Protection."

    async def generate_csrf_token(self, client_ip: str) -> str:
        token = secrets.token_hex(32)
        csrf_data = {"token": token, "ip": client_ip, "_created": datetime.utcnow().timestamp()}
        csrf_token = await asyncio.to_thread(self.serializer.dumps, csrf_data, self.secret_key)
        self.logger.info(f"CSRF token generated for IP {client_ip}")
        return csrf_token

    async def inject_csrf_token(self, response: Response, csrf_token: str) -> Response:
        try:
            await response.set_cookie(
                self.csrf_token_key,
                csrf_token,
                max_age=self.expiration.total_seconds() if self.expiration else None,
                **self.cookie_options
            )
            self.logger.info("CSRF token injected into the response")
            return response
        except Exception as e:
            self.logger.error(f"Failed to inject CSRF token: {e}")
            raise CSRFTokenError("Failed to inject CSRF token")

    async def validate_csrf_token(self, request: Request) -> bool:
        try:
            csrf_token = request.cookies.get(self.csrf_token_key)
            form = await request.form()
            submitted_token = form["_csrf_token"]

            current_time = await asyncio.to_thread(datetime.utcnow)
            if self.max_token_lifetime and not await self._is_within_lifetime(csrf_token, current_time):
                self.logger.warning("CSRF token validation failed: Token has expired.")
                return False

            if csrf_token in self.revoked_tokens:
                self.logger.warning("CSRF token validation failed: Token has been revoked.")
                return False

            decoded_csrf_token = await asyncio.to_thread(self.serializer.loads, csrf_token)
            decoded_submitted_token = await asyncio.to_thread(self.serializer.loads, submitted_token, self.secret_key)

            if self.ip_verification and decoded_csrf_token.get("ip") != request.client.host:
                self.logger.warning("CSRF token validation failed: IP address mismatch.")
                return False

            if self.trusted_ips and '*' not in self.trusted_ips and request.client.host not in self.trusted_ips:
                self.logger.warning("CSRF token validation failed: Untrusted IP address.")
                return False

            return csrf_token is not None and decoded_csrf_token == decoded_submitted_token
        except (signing.BadSignature, signing.SignatureExpired) as e:
            self.logger.warning(f"CSRF token validation failed: {e}")
            return False
        except Exception as e:
            self.logger.error(f"Error validating CSRF token: {e}")
            raise CSRFTokenError("Error validating CSRF token")

    async def _is_within_lifetime(self, csrf_token: str, current_time: datetime) -> bool:
        if not self.max_token_lifetime:
            return True

        try:
            token_data = await asyncio.to_thread(self.serializer.loads, csrf_token, self.secret_key)
            created_time = datetime.fromtimestamp(token_data["_created"])
            return (current_time - created_time) <= self.max_token_lifetime
        except Exception as e:
            self.logger.warning(f"Failed to extract creation time from CSRF token: {e}")
            return False

    async def should_refresh_csrf_token(self, csrf_token: str) -> bool:
        if not self.token_refresh_interval:
            return False

        try:
            token_data = await asyncio.to_thread(self.serializer.loads, csrf_token, self.secret_key)
            created_time = datetime.fromtimestamp(token_data["_created"])
            return (datetime.utcnow() - created_time) >= self.token_refresh_interval
        except Exception as e:
            self.logger.warning(f"Failed to extract creation time from CSRF token: {e}")
            return False

    def get_client_ip(self, request: Request) -> str:
        return request.headers.get("X-Real-IP") or request.client.host or "127.0.0.1"

    def revoke_csrf_token(self, csrf_token: str):
        self.revoked_tokens.add(csrf_token)



from .sqlite3 import (
    Session as Session,
    Sqlite3 as Sqlite3,
    create_session as create_session
)

LocalSessionManager = create_session()

connection = Session

__all__ = [
    LocalSessionManager,
    connection,
    Sqlite3
]


from .localsessions import SessionManager as LocalSessionManager
from .inmemorysession import InMemorySessionBeforeStage as InMemorySessionBeforeStage
from .cookiesession import BeforeSessionStage as BeforeCookieSessionStage, AfterSessionStage as AfterCookieSessionStage

__all__ = [
    "LocalSessionManager",
    "InMemorySessionBeforeStage",
    "BeforeCookieSessionStage",
    "AfterCookieSessionStage"
]


import importlib
import os
import traceback
import warnings
import aquilify
from . import global_settings

from aquilify.utils.functional import LazyObject, empty

class ImproperlyConfigured(Exception):
    pass

ENVIRONMENT_VARIABLE = "AQUILIFY_SETTINGS_MODULE"

class SettingsReference(str):
    def __new__(self, value, setting_name):
        return str.__new__(self, value)
    def __init__(self, value, setting_name):
        self.setting_name = setting_name

class LazySettings(LazyObject):
    def _setup(self, name=None):
        settings_module = os.environ.get(ENVIRONMENT_VARIABLE)
        if not settings_module:
            desc = ("setting %s" % name) if name else "settings"
            raise ImproperlyConfigured(
                "Requested %s, but settings are not configured. "
                "You must either define the environment variable %s "
                "or call settings.configure() before accessing settings."
                % (desc, ENVIRONMENT_VARIABLE)
            )
        self._wrapped = Settings(settings_module)
    def __repr__(self):
        if self._wrapped is empty:
            return "<LazySettings [Unevaluated]>"
        return '<LazySettings "%(settings_module)s">' % {
            "settings_module": self._wrapped.SETTINGS_MODULE,
        }
    def __getattr__(self, name):
        if (_wrapped := self._wrapped) is empty:
            self._setup(name)
            _wrapped = self._wrapped
        val = getattr(_wrapped, name)
        if name == "SECRET_KEY" and not val:
            raise ImproperlyConfigured("The SECRET_KEY setting must not be empty.")
        self.__dict__[name] = val
        return val
    def __setattr__(self, name, value):
        if name == "_wrapped":
            self.__dict__.clear()
        else:
            self.__dict__.pop(name, None)
        super().__setattr__(name, value)
    def __delattr__(self, name):
        super().__delattr__(name)
        self.__dict__.pop(name, None)
    def configure(self, default_settings=global_settings, **options):
        if self._wrapped is not empty:
            raise RuntimeError("Settings already configured.")
        holder = UserSettingsHolder(default_settings)
        for name, value in options.items():
            if not name.isupper():
                raise TypeError("Setting %r must be uppercase." % name)
            setattr(holder, name, value)
        self._wrapped = holder
    @property
    def configured(self):
        return self._wrapped is not empty
    def _show_deprecation_warning(self, message, category):
        stack = traceback.extract_stack()
        filename, _, _, _ = stack[-4]
        if not filename.startswith(os.path.dirname(aquilify.__file__)):
            warnings.warn(message, category, stacklevel=2)

class Settings:
    def __init__(self, settings_module):
        for setting in dir(global_settings):
            if setting.isupper():
                setattr(self, setting, getattr(global_settings, setting))
        self.SETTINGS_MODULE = settings_module
        mod = importlib.import_module(self.SETTINGS_MODULE)
        tuple_settings = (
            "ALLOWED_HOSTS",
            "TEMPLATES",
            "SECRET_KEY_FALLBACKS",
        )
        self._explicit_settings = set()
        for setting in dir(mod):
            if setting.isupper():
                setting_value = getattr(mod, setting)
                if setting in tuple_settings and not isinstance(setting_value, (list, tuple)):
                    raise ImproperlyConfigured(
                        "The %s setting must be a list or a tuple." % setting
                    )
                setattr(self, setting, setting_value)
                self._explicit_settings.add(setting)
                
    def is_overridden(self, setting):
        return setting in self._explicit_settings

    def __repr__(self):
        return '<%(cls)s "%(settings_module)s">' % {
            "cls": self.__class__.__name__,
            "settings_module": self.SETTINGS_MODULE,
        }

class UserSettingsHolder:

    SETTINGS_MODULE = None

    def __init__(self, default_settings):
        self.__dict__["_deleted"] = set()
        self.default_settings = default_settings

    def __getattr__(self, name):
        if not name.isupper() or name in self._deleted:
            raise AttributeError
        return getattr(self.default_settings, name)

    def __setattr__(self, name, value):
        self._deleted.discard(name)
        super().__setattr__(name, value)

    def __delattr__(self, name):
        self._deleted.add(name)
        if hasattr(self, name):
            super().__delattr__(name)

    def __dir__(self):
        return sorted(
            s
            for s in [*self.__dict__, *dir(self.default_settings)]
            if s not in self._deleted
        )

    def is_overridden(self, setting):
        deleted = setting in self._deleted
        set_locally = setting in self.__dict__
        set_on_default = getattr(
            self.default_settings, "is_overridden", lambda s: False
        )(setting)
        return deleted or set_locally or set_on_default

    def __repr__(self):
        return "<%(cls)s>" % {
            "cls": self.__class__.__name__,
        }


settings = LazySettings()



import importlib

class ASGIMiddlewareLoaderError(Exception):
    pass

class ASGIMiddlewareLoader:
    def __init__(self, settings_module='settings'):
        self.settings_module = settings_module
        
    def load_asgi_middlewares(self):
        try:
            settings = importlib.import_module(self.settings_module)
            asgi_middlewares = getattr(settings, 'ASGI_MIDDLEWARES', [])
            
            if not asgi_middlewares:
                pass

            if not isinstance(asgi_middlewares, list):
                raise ASGIMiddlewareLoaderError("Invalid ASGI_MIDDLEWARES configuration")

            middlewares = []

            for middleware_path in asgi_middlewares:
                module_name, object_name = middleware_path.rsplit('.', 1)
                middleware_module = importlib.import_module(module_name)
                middleware = getattr(middleware_module, object_name)
                middlewares.append(middleware)

            return middlewares

        except ImportError:
            raise ASGIMiddlewareLoaderError(f"Could not import {self.settings_module}.")
        except AttributeError:
            raise ASGIMiddlewareLoaderError(f"ASGI_MIDDLEWARES not found in {self.settings_module}.")
        except ASGIMiddlewareLoaderError as e:
            raise e
        except Exception as e:
            raise ASGIMiddlewareLoaderError(f"Error: {e}")
